import traceback as tb
import datetime
import a99

__all__ = ["_LogPart"]

class _LogPart(object):
    """
    This class is a base for XLogDialog and XLogMainWindow
    """
    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Interface

    def keep_ref(self, obj):
        """Adds obj to internal list to keep a reference to it.

        When using PyQt, it happens that the Python object gets garbage-collected even
        when a C++ Qt object still exists, causing a mess
        """
        self._refs.append(obj)
        return obj


    def add_log_error(self, x, flag_also_show=False, E=None):
        """Sets text of labelError."""
        if len(x) == 0:
            x = "(empty error)"
            tb.print_stack()
        x_ = x
        if E is not None:
            a99.get_python_logger().exception(x_)
        else:
            a99.get_python_logger().info("ERROR: {}".format(x_))

        x = '<span style="color: {0!s}">{1!s}</span>'.format(a99.COLOR_ERROR, x)
        self._add_log_no_logger(x, False)
        if flag_also_show:
            a99.show_error(x_)

    def add_log(self, x, flag_also_show=False):
        """Logs to 4 different outputs: conditionally to 3, and certainly to get_python_logger()"""

        self._add_log_no_logger(x, flag_also_show)

        a99.get_python_logger().info(x)

    def _add_log_no_logger(self, x, flag_also_show):
        if hasattr(self, "label_last_log"):
            self.label_last_log.setText(x)
        if hasattr(self, "textEdit_log"):
            te = self.textEdit_log
            te.append("{0!s} -- {1!s}".format(a99.now_str(), x))
        if flag_also_show:
            a99.show_error(x)

    def status(self, x):
        """Sets text of label_last_log"""
        if hasattr(self, "label_last_log"):
            self.label_last_log.setText(x)

    def clear_log(self):
        """Clears text of label_last_log"""
        if hasattr(self, "label_last_log"):
            self.label_last_log.setText("")



"""Logging routines"""


__all__ = [
"get_python_logger", "add_file_handler",
"LogTwo", "SmartFormatter", "str_exc"]


import logging
import sys
from argparse import *
from .parts import *
import a99
import datetime
import traceback



_python_logger = None
_fmtr = logging.Formatter('[%(levelname)-8s] %(message)s')
def get_python_logger():
    """Returns logger to receive Python messages (as opposed to Fortran).

    Note: we don't want this to be called at any module imports, in order to give the API user a
          chance to set `a99.flag_log_file` and `a99.flag_log_console`
    """
    global _python_logger
    if _python_logger is None:
        fn = "python.log"
        l = logging.Logger("python", level=a99.logging_level)
        if a99.flag_log_file:
            add_file_handler(l, fn)
        if a99.flag_log_console:
            ch = logging.StreamHandler()
            ch.setFormatter(_fmtr)
            l.addHandler(ch)
        _python_logger = l
        for line in a99.format_box("Session started @ {}".format(a99.now_str())):
            l.info(line)
        if a99.flag_log_file:
            l.info("$ Logging to console $")
        if a99.flag_log_file:
            l.info("$ Logging to file '{}' $".format(fn))

    return _python_logger


def add_file_handler(logger, logFilename=None):
    """Adds file handler to logger.

    File is opened in "a" mode (append)
    """
    assert isinstance(logger, logging.Logger)
    ch = logging.FileHandler(logFilename, "a")
    # ch.setFormatter(logging._defaultFormatter) # todo may change to have same formatter as last handler of logger
    ch.setFormatter(_fmtr)
    logger.addHandler(ch)


@froze_it
class LogTwo(object):
  """Logs messages to both stdout and file."""
  def __init__(self, filename):
    self.terminal = sys.stdout
    self.log = open(filename, "w")

  def write(self, message):
    self.terminal.write(message)
    self.log.write(message)

  def close(self):
      self.log.close()


class SmartFormatter(RawDescriptionHelpFormatter):
    """
    Help formatter that will show default option values and also respect
    newlines in description. Neither are done in default help formatter.
    """

    def _get_help_string(self, action):
        help = action.help
        if '%(default)' not in action.help:
            if action.default is not SUPPRESS:
                defaulting_nargs = [OPTIONAL, ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += ' (default: %(default)s)'
        return help


        # # this is the RawTextHelpFormatter._split_lines
        # if text.startswith('R|'):
        #     return text[2:].splitlines()
        # return argparse.ArgumentDefaultsHelpFormatter._split_lines(self, text, width)


def str_exc(E):
    """Generates a string from an Exception"""
    return "{0!s}: {1!s}".format(E.__class__.__name__, str(E))



"""A holed layer over the sqlite3 API"""

__all__ = ["get_conn", "conn_is_open", "cursor_to_data_header", "cursor_to_rows", "MyDBRow",
           "TableInfo", "get_table_names", "get_table_info"]

from collections import OrderedDict
import sqlite3


def get_conn(filename):
    """Returns new sqlite3.Connection object with _dict_factory() as row factory"""
    conn = sqlite3.connect(filename)
    conn.row_factory = _dict_factory
    return conn


def conn_is_open(conn):
    """Tests sqlite3 connection, returns T/F"""
    if conn is None:
        return False

    try:
        get_table_names(conn)
        return True

        # # Idea taken from
        # # http: // stackoverflow.com / questions / 1981392 / how - to - tell - if -python - sqlite - database - connection - or -cursor - is -closed
        # conn.execute("select id from molecule limit 1")
        # return True
    except sqlite3.ProgrammingError as e:
        # print(e)
        return False

def get_table_names(conn):
    # http://stackoverflow.com/questions/305378/list-of-tables-db-schema-dump-etc-using-the-python-sqlite3-api

    r = conn.execute("select name from sqlite_master where type = 'table'")
    names = [row.name for row in r]
    return names


def cursor_to_data_header(cursor):
    """Fetches all rows from query ("cursor") and returns a pair (data, header)

    Returns: (data, header), where
        - data is a [num_rows]x[num_cols] list of lists;
        - header is a [num_cols] list containing the field names
    """
    n = 0
    data, header = [], {}
    for row in cursor:
        if n == 0:
            header = row.keys()
        data.append(row.values())
    return data, list(header)


def cursor_to_rows(cursor):
    """Fetches all rows from query ("cursor") and returns a list of MyDBRow objects
    """
    ret = []
    for row in cursor:
        ret.append(row)
    return ret


class MyDBRow(OrderedDict):
    """
    Dict subclass that allows attribute-like access of values using keys
    """
    def __getattribute__(self, name):
        """
        Allows attribute-like access of dictionary values
        """
        if name in self:
            return self[name]
        return OrderedDict.__getattribute__(self, name)
        # raise AttributeError("'{}' object has no attribute '{}' #".format(self.__class__.__name__, name))

    def None_to_zero(self):
        """Replace None with zero"""

        for key in self:
            if self[key] is None:
                self[key] = 0.


class TableInfo(OrderedDict):
    """
    Represents the information about a table in dict-like form where key is the field name
    """
    def find(self, **kwargs):
        """
        Finds row matching specific field value

        Args:
            **kwargs: (**only one argument accepted**) fielname=value, e.g., formula="OH"

        Returns: list element or None
        """

        if len(kwargs) != 1:
            raise ValueError("One and only one keyword argument accepted")

        key = list(kwargs.keys())[0]
        value = list(kwargs.values())[0]
        ret = None
        for row in self.values():
            if row[key] == value:
                ret = row
                break
        return ret


def get_table_info(conn, tablename):
    """Returns TableInfo object"""
    r = conn.execute("pragma table_info('{}')".format(tablename))
    ret = TableInfo(((row["name"], row) for row in r))
    return ret


# http://stackoverflow.com/questions/3300464/how-can-i-get-dict-from-sqlite-query
def _dict_factory(cursor, row):
    d = MyDBRow()
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d




__all__ = ["GetCurrentFunctionName", "MyLock"]


import inspect
from threading import Lock, current_thread
import traceback
import a99


###############################################################################
# # Debugging facilities

def GetCurrentFunctionName(k=0):
  """Returns the function name of the caller.

  Arguments:
    k=0 -- 1: caller of the caller,
       2: caller of the caller of the caller etc

  Obs: profiling shows that this function grabs a lot of execution time for itself."""
  return inspect.stack()[1+k][3]


class MyLock(object):
    """Lock with verbosing, helps to find deadlocks"""

    def __init__(self, name=None, flag_verbose=False):
        if name is None:
            name = a99.random_name()
        self.name = name
        self.flag_verbose = flag_verbose
        self.__lock = Lock()

    def acquire(self, *args):
        if self.__lock.locked():
            self.__log("Tried to lock, but is already locked!!")
            a99.get_python_logger().info("\n".join(traceback.format_stack()))
        else:
            pass
            # self.__log("Will acquire lock")
        self.__lock.acquire(*args)
        self.__log("Acquired lock, good")

    def release(self):
        self.__lock.release()
        self.__log("Released lock")

    # For the "with" statement
    def __exit__(self, *args):
        self.release()
    __enter__ = acquire

    def __log(self, s):
        if self.flag_verbose:
            a99.get_python_logger().info("--- MyLock {0!s} --- {1!s} (caller: {2!s}; thread: {3!s})".format(self.name, s, GetCurrentFunctionName(2), current_thread().name))




__all__ = ["WCollapsiblePanel"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99


class WCollapsiblePanel(QWidget):
    """
    Collapsible panel

    Attributes:
       self.widget -- populate this
       self.collapsed_label -- label to become visible when collapsed. Set its text to show some
                               text when collapsed
    """

    def __init__(self, *args):
        QWidget.__init__(self, *args)

        self.flag_collapsed = False
        self.button_text_expanded = "^"
        self.button_text_collapsed = "v"


        lh = self.layout = QHBoxLayout(self)
        a99.set_margin(lh, 0)
        lh.setSpacing(2)

        lv = self.layout1 = QVBoxLayout()
        lh.addLayout(lv)
        a99.set_margin(lv, 0)
        lv.setSpacing(0)

        b = self.button = QPushButton("^")
        b.setFixedWidth(16)
        b.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        b.clicked.connect(self.collapse_expand_clicked)
        lv.addWidget(b)
        lv.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        self.widget = QWidget()
        lh.addWidget(self.widget)

        # lh.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        label = self.label = QLabel()
        label.setVisible(False)
        lh.addWidget(label)

    def collapse_expand_clicked(self):
        self.flag_collapsed =  not self.flag_collapsed
        self.widget.setVisible(not self.flag_collapsed)
        self.button.setText(self.button_text_collapsed if self.flag_collapsed else self.button_text_expanded)
        self.label.setVisible(self.flag_collapsed)



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget
from .a_XLogDialog import XLogDialog
from .a_XLogMainWindow import XLogMainWindow

__all__ = ["WBase"]


class WBase(QWidget):
    """Widget with 'changed' signal, keep_ref(), logging tools"""

    # Should be emitted whenever any value changes
    changed = pyqtSignal()


    def __init__(self, parent):
        assert isinstance(parent, (XLogMainWindow, XLogDialog))
        QWidget.__init__(self, parent)
        self._refs = []
        self.parent_form = parent


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Interface

    def keep_ref(self, obj):
        """Adds obj to internal list to keep a reference to it.

        WHen using PyQt, it happens that the Python object gets garbage-collected even
        when a C++ Qt object still exists, causing a mess
        """
        self._refs.append(obj)
        return obj

    def clear_log(self):
        self.parent_form.clear_log()

    def add_log_error(self, x, flag_also_show=False, E=None):
        """Delegates to parent form"""
        self.parent_form.add_log_error(x, flag_also_show, E)

    def add_log(self, x, flag_also_show=False):
        """Delegates to parent form"""
        self.parent_form.add_log(x, flag_also_show)

    def status(self, msg):
        """Probably shows message in status bar"""
        self.parent_form.status(msg)

    def update_gui_label_fn(self):
        pass






import textwrap
import sys

__all__ = ["format_h1", "format_h2", "fmt_error", "print_error", "menu", "format_progress", "markdown_table",
           "format_exe_info", "format_box", "yesno", "rest_table", "expand_multirow_data"]


NIND = 2  # Number of spaces per indentation level

# #################################################################################################
# # Text interface routines - routines that are useful for building a text interface


def format_underline(s, char="=", indents=0):
    """
    Traces a dashed line below string

    Args:
        s: string
        intents: number of leading intenting spaces
        format: string starting with "text" or "markdown"

    Returns: list

    >>> print("\\n".join(format_underline("Life of João da Silva", "^", 2)))
      Life of João da Silva
      ^^^^^^^^^^^^^^^^^^^^^
    """

    n = len(s)
    ind = " " * indents
    return ["{}{}".format(ind, s), "{}{}".format(ind, char*n)]


def format_h1(s, format="text", indents=0):
    """
    Encloses string in format text

    Args:
        s: string
        intents: number of leading intenting spaces
        format: string starting with "text", "markdown", or "rest"

    Returns: list

    >>> print("\\n".join(format_h2("Header 1", indents=10)))
              Header 1
              --------

    >>> print("\\n".join(format_h2("Header 1", "markdown", 0)))
    ## Header 1
    """

    if format.startswith("text"):
        return format_underline(s, "=", indents)
    elif format.startswith("markdown"):
        return ["# {}".format(s)]
    elif format.startswith("rest"):
        return format_underline(s, "=", 0)


def format_h2(s, format="text", indents=0):
    """
    Encloses string in format text

    Args:
        s: string
        format: string starting with "text" or "markdown"
        intents: number of leading intenting spaces

    Returns: list

    >>> print("\\n".join(format_h2("Header 2", indents=2)))
      Header 2
      --------

    >>> print("\\n".join(format_h2("Header 2", "markdown", 2)))
    ## Header 2
    """

    if format.startswith("text"):
        return format_underline(s, "-", indents)
    elif format.startswith("markdown"):
        return ["## {}".format(s)]
    elif format.startswith("rest"):
        return format_underline(s, "-", 0)



def fmt_error(s):
    """Standardized embellishment. Adds formatting to error message."""
    return "!! {0!s} !!".format(s)


def print_error(s):
    """Prints string as error message."""
    print((fmt_error(s)))


def yesno(question, default=None):
    """Asks a yes/no question

    Args:
        question: string **without** the question mark and without the options.
            Example: 'Create links'
        default: default option. Accepted values are 'Y', 'YES', 'N', 'NO' or lowercase versions of
            these valus (this argument is case-insensitive)

    Returns:
        bool: True if user answered Yes, False otherwise
    """

    if default is not None:
        if isinstance(default, bool):
            pass
        else:
            default_ = default.upper()
            if default_ not in ('Y', 'YES', 'N', 'NO'):
                raise RuntimeError("Invalid default value: '{}'".format(default))
            default = default_ in ('Y', 'YES')

    while True:
        ans = input("{} ({}/{})? ".format(question, "Y" if default == True else "y",
                                         "N" if default == False else "n")).upper()
        if ans == "" and default is not None:
            ret = default
            break
        elif ans in ("N", "NO"):
            ret = False
            break
        elif ans in ("Y", "YES"):
            ret = True
            break
    return ret


def menu(title, options, cancel_label="Cancel", flag_allow_empty=False, flag_cancel=True, ch='.'):
  """Text menu.

  Arguments:
    title -- menu title, to appear at the top
    options -- sequence of strings
    cancel_label='Cancel' -- label to show at last "zero" option
    flag_allow_empty=0 -- Whether to allow empty option
    flag_cancel=True -- whether there is a "0 - Cancel" option
    ch="." -- character to use to draw frame around title

  Returns:
    option -- an integer: None; 0-Back/Cancel/etc; 1, 2, ...

  Adapted from irootlab menu.m"""

  num_options, flag_ok = len(options), 0
  option = None  # result
  min_allowed = 0 if flag_cancel else 1  # minimum option value allowed (if option not empty)

  while True:
    print("")
    for line in format_box(title, ch):
        print("  "+line)
    for i, s in enumerate(options):
      print(("  {0:d} - {1!s}".format(i+1, s)))
    if flag_cancel: print(("  0 - << (*{0!s}*)".format(cancel_label)))
    try:
        s_option = input('? ')
    except KeyboardInterrupt:
        raise
    except:
        print("")

    n_try = 0
    while True:
      if n_try >= 10:
        print('You are messing up!')
        break

      if len(s_option) == 0 and flag_allow_empty:
        flag_ok = True
        break

      try:
        option = int(s_option)
        if min_allowed <= option <= num_options:
          flag_ok = True
          break
      except ValueError:
        print("Invalid integer value!")

      print(("Invalid option, range is [{0:d}, {1:d}]!".format(0 if flag_cancel else 1, num_options)))

      n_try += 1
      s_option = input("? ")

    if flag_ok:
      break
  return option


def format_box(title, ch="*"):
    """
    Encloses title in a box. Result is a list

    >>> for line in format_box("Today's TODO list"):
    ...     print(line)
    *************************
    *** Today's TODO list ***
    *************************
    """
    lt = len(title)
    return [(ch * (lt + 8)),
            (ch * 3 + " " + title + " " + ch * 3),
            (ch * (lt + 8))
           ]


def format_progress(i, n):
    """Returns string containing a progress bar, a percentage, etc."""
    if n == 0:
        fraction = 0
    else:
        fraction = float(i)/n
    LEN_BAR = 25
    num_plus = int(round(fraction*LEN_BAR))
    s_plus = '+'*num_plus
    s_point = '.'*(LEN_BAR-num_plus)
    return '[{0!s}{1!s}] {2:d}/{3:d} - {4:.1f}%'.format(s_plus, s_point, i, n, fraction*100)


# #################################################################################################
# # Formatting for listing the programs available


def _format_exe_info(py_len, exeinfo, format, indlevel):
    """Renders ExeInfo object in specified format"""
    ret = []
    ind = " " * indlevel * NIND if format.startswith("text") else ""
    if format == "markdown-list":
        for si in exeinfo:
            ret.append("  - `{0!s}`: {1!s}".format(si.filename, si.description))
    if format == "rest-list":
        for si in exeinfo:
            ret.append("* ``{0!s}``: {1!s}".format(si.filename, si.description))
    elif format == "markdown-table":
        mask = "%-{0:d}s | %s".format(py_len+2 )
        ret.append(mask % ("Script name", "Purpose"))
        ret.append("-" * (py_len + 3) + "|" + "-" * 10)
        for si in exeinfo:
            ret.append(mask % ("`{0!s}`".format(si.filename), si.description))
    elif format == "text":
        sbc = 1  # spaces between columns
        for si in exeinfo:
            ss = textwrap.wrap(si.description, 79 - py_len - sbc - indlevel*NIND)
            for i, s in enumerate(ss):
                if i == 0:
                    filecolumn = si.filename + " " + ("." * (py_len - len(si.filename)))
                else:
                    filecolumn = " " * (py_len + 1)

                ret.append("{}{}{}{}".format(ind, filecolumn, " "*sbc, s))
    ret.append("")
    return ret


def format_exe_info(exeinfo, format="text", indlevel=0):
    """
    Generates listing of all Python scripts available as command-line programs.

    Args:
      exeinfo -- list of ExeInfo objects

      format -- One of the options below:
        "text" -- generates plain text for printing at the console
        "markdown-list" -- generates MarkDown as list
        "markdown-table" -- generates MarkDown as tables
        "rest-list" -- generates reStructuredText as lists

      indents -- indentation level ("text" format only)

    Returns: (list of strings, maximum filename size)
      list of strings -- can be joined with a "\n"
      maximum filename size
    """

    py_len = max([len(si.filename) for si in exeinfo])

    sisi_gra = [si for si in exeinfo if si.flag_gui == True]
    sisi_cmd = [si for si in exeinfo if si.flag_gui == False]
    sisi_none = [si for si in exeinfo if si.flag_gui is None]

    def get_title(x):
        return format_h2(x, format, indlevel*NIND) + [""]

    ret = []
    if len(sisi_gra) > 0:
        ret.extend(get_title("Graphical applications"))
        ret.extend(_format_exe_info(py_len, sisi_gra, format, indlevel + 1))
    if len(sisi_cmd) > 0:
        ret.extend(get_title("Command-line tools", ))
        ret.extend(_format_exe_info(py_len, sisi_cmd, format, indlevel + 1))
    if len(sisi_none) > 0:
        ret.extend(_format_exe_info(py_len, sisi_none, format, indlevel + 1))

    return ret, py_len


# #################################################################################################
# # Text table functions

def markdown_table(data, headers):
    """
    Creates MarkDown table. Returns list of strings

    Arguments:
      data -- [(cell00, cell01, ...), (cell10, cell11, ...), ...]
      headers -- sequence of strings: (header0, header1, ...)
    """

    maxx = [max([len(x) for x in column]) for column in zip(*data)]
    maxx = [max(ll) for ll in zip(maxx, [len(x) for x in headers])]
    mask = " | ".join(["%-{0:d}s".format(n) for n in maxx])

    ret = [mask % headers]

    ret.append(" | ".join(["-"*n for n in maxx]))
    for line in data:
        ret.append(mask % line)
    return ret


def expand_multirow_data(data):
    """
    Converts multirow cells to a list of lists and informs the number of lines of each row.

    Returns:
         tuple: new_data, row_heights
    """

    num_cols = len(data[0]) # number of columns

    # calculates row heights
    row_heights = []
    for mlrow in data:
        row_height = 0
        for j, cell in enumerate(mlrow):
            row_height = max(row_height, 1 if not isinstance(cell, (list, tuple)) else len(cell))
        row_heights.append(row_height)
    num_lines = sum(row_heights) # line != row (rows are multiline)

    # rebuilds table data
    new_data = [[""]*num_cols for i in range(num_lines)]
    i0 = 0
    for row_height, mlrow in zip(row_heights, data):
        for j, cell in enumerate(mlrow):
            if not isinstance(cell, (list, tuple)):
                cell = [cell]

            for incr, x in enumerate(cell):
                new_data[i0+incr][j] = x

        i0 += row_height

    return new_data, row_heights



def rest_table(data, headers):
    """
    Creates reStructuredText table (grid format), allowing for multiline cells

    Arguments:
      data -- [((cell000, cell001, ...), (cell010, cell011, ...), ...), ...]
      headers -- sequence of strings: (header0, header1, ...)

    **Note** Tolerant to non-strings

    **Note** Cells may or may not be multiline

    >>> rest_table([["Eric", "Idle"], ["Graham", "Chapman"], ["Terry", "Gilliam"]], ["Name", "Surname"])
    """

    num_cols = len(headers)
    new_data, row_heights = expand_multirow_data(data)
    new_data = [[str(x) for x in row] for row in new_data]
    col_widths = [max([len(x) for x in col]) for col in zip(*new_data)]
    col_widths = [max(cw, len(s)) for cw, s in zip(col_widths, headers)]

    if any([x == 0 for x in col_widths]):
        raise RuntimeError("Column widths ({}) has at least one zero".format(col_widths))

    num_lines = sum(row_heights) # line != row (rows are multiline)

    # horizontal lines
    hl0 = "+"+"+".join(["-"*(n+2) for n in col_widths])+"+"
    hl1 = "+"+"+".join(["="*(n+2) for n in col_widths])+"+"

    frmtd = ["{0:{1}}".format(x, width) for x, width in zip(headers, col_widths)]
    ret = [hl0, "| "+" | ".join(frmtd)+" |", hl1]

    i0 = 0
    for row_height in row_heights:
        for incr in range(row_height):
            frmtd = ["{0:{1}}".format(x, width) for x, width in zip(new_data[i0+incr], col_widths)]
            ret.append("| "+" | ".join(frmtd)+" |")
        i0 += row_height

    ret.append(hl0)
    return ret



"""Conversion routines"""


__all__ = [
"str2bool", "bool2str", "chunk_string", "ordinal_suffix", "seconds2str", "make_fits_keys_dict",
"valid_fits_key", "eval_fieldnames", "expr_to_fieldname", "module_to_dict", "unicode_to_greek",
"greek_to_unicode", "make_code_readable", "int_to_superscript"]


import numpy as np
import re


# TODO maybe a str2bool_ex() and its counterpart

def str2bool(s):
    """Understands "T"/"F" only (case-sensitive). To be used for file parsing.

    **Note** This routine is limited on purpose for speed.
    """
    if s == "T":
        return True
    elif s == "F":
        return False
    raise ValueError("I don't understand '{0!s}' as a logical value".format(s))


def bool2str(x):
    """Converts bool variable to either "T" or "F".

    **Note** This routine is limited on purpose for speed.
    """
    assert isinstance(x, bool)
    return "T" if x else "F"


def make_code_readable(s):
    """Adds newlines at strategic places"""

    MAP = {",": ",\n", "{": "{\n ", "}": "\n}"}

    ll = []

    state = "open"
    flag_single = False
    flag_double = False
    flag_backslash = False
    for ch in s:
        if flag_backslash:
            flag_backslash = False
            continue

        if ch == "\\":
            flag_backslash = True
            continue

        if flag_single:
            if ch == "'":
                flag_single = False
        elif not flag_double and ch == "'":
            flag_single = True

        if flag_double:
            if ch == '"':
                flag_double = False
        elif not flag_single and ch == '"':
            flag_double = True

        if flag_single or flag_double:
            ll.append(ch)
        else:
            ll.append(MAP.get(ch, ch))

    return "".join(ll)


def chunk_string(string, length):
    """
    Splits a string into fixed-length chunks.

    This function returns a generator, using a generator comprehension. The
    generator returns the string sliced, from 0 + a multiple of the length
    of the chunks, to the length of the chunks + a multiple of the length
    of the chunks.

    Reference: http://stackoverflow.com/questions/18854620
    """
    return (string[0 + i:length + i] for i in range(0, len(string), length))


def ordinal_suffix(i):
    """Returns 'st', 'nd', or 'rd'."""
    return 'st' if i == 1 else 'nd' if i == 2 else 'rd' if i == 3 else 'th'


def seconds2str(seconds):
    """Returns string such as 1h 05m 55s."""

    if seconds < 0:
        return "{0:.3g}s".format(seconds)
    elif np.isnan(seconds):
        return "NaN"
    elif np.isinf(seconds):
        return "Inf"

    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    if h >= 1:
        return "{0:g}h {1:02g}m {2:.3g}s".format(h, m, s)
    elif m >= 1:
        return "{0:02g}m {1:.3g}s".format(m, s)
    else:
        return "{0:.3g}s".format(s)


def make_fits_keys_dict(keys):
    """
    Returns a dictionary to translate to unique FITS header keys up to 8 characters long

    This is similar to Windows making up 8-character names for filenames that
    are longer than this

    "The keyword names may be up to 8 characters long and can only contain
    uppercase letters A to Z, the digits 0 to 9, the hyphen, and the underscore
    character." [1]

    Arguments:
        keys -- list of strings

    Returns:
        dictionary whose keys are the elements in the "keys" argument, and whose
        values are made-up uppercase names

    References:
        [1] http://fits.gsfc.nasa.gov/fits_primer.html
    """

    key_dict = {}
    new_keys = []
    for key in keys:
        # converts to valid FITS key according to reference [1] above
        fits_key = valid_fits_key(key)
        num_digits = 1
        i = -1
        i_max = 9
        while fits_key in new_keys:
            i += 1
            if i > i_max:
                i = 0
                i_max = i_max * 10 + 9
                num_digits += 1
            fits_key = fits_key[:(8 - num_digits)] + (("%0{0:d}d".format(num_digits)) % i)

        key_dict[key] = fits_key
        new_keys.append(fits_key)

    return key_dict


def valid_fits_key(key):
    """
    Makes valid key for a FITS header

    "The keyword names may be up to 8 characters long and can only contain
    uppercase letters A to Z, the digits 0 to 9, the hyphen, and the underscore
    character." (http://fits.gsfc.nasa.gov/fits_primer.html)
    """

    ret = re.sub("[^A-Z0-9\-_]", "", key.upper())[:8]
    if len(ret) == 0:
        raise RuntimeError("key '{0!s}' has no valid characters to be a key in a FITS header".format(key))
    return ret


def eval_fieldnames(string_, varname="fieldnames"):
    """Evaluates string_, must evaluate to list of strings. Also converts field names to uppercase"""
    ff = eval(string_)
    if not isinstance(ff, list):
        raise RuntimeError("{0!s} must be a list".format(varname))
    if not all([isinstance(x, str) for x in ff]):
        raise RuntimeError("{0!s} must be a list of strings".format(varname))
    ff = [x.upper() for x in ff]
    return ff


def expr_to_fieldname(expr):
    """Keeps the rightmost part of the class name after the "_" """
    return expr[:expr.index("(")].strip().split("_")[-1]


def module_to_dict(module):
    """Creates a dictionary whose keys are module.__all__

    Returns: {"(attribute name)": attribute, ...}
    """

    lot = [(key, module.__getattribute__(key)) for key in module.__all__]
    ret = dict(lot)
    return ret



# **        ****                ******        ****                ******        ****
#   **    **    ******    ******      **    **    ******    ******      **    **    ******    ******
#     ****            ****              ****            ****              ****            ****
#
# Greek alphabet-related routines

# Source:
#     "A Python dictionary mapping the Unicode codes of the greek alphabet to their names"
#     https://gist.github.com/beniwohli/765262
#
_UNICODE_GREEK = (
('\u0391', 'Alpha'),
('\u0392', 'Beta'),
('\u0393', 'Gamma'),
('\u0394', 'Delta'),
('\u0395', 'Epsilon'),
('\u0396', 'Zeta'),
('\u0397', 'Eta'),
('\u0398', 'Theta'),
('\u0399', 'Iota'),
('\u039A', 'Kappa'),
('\u039B', 'Lamda'),
('\u039C', 'Mu'),
('\u039D', 'Nu'),
('\u039E', 'Xi'),
('\u039F', 'Omicron'),
('\u03A0', 'Pi'),
('\u03A1', 'Rho'),
('\u03A3', 'Sigma'),
('\u03A4', 'Tau'),
('\u03A5', 'Upsilon'),
('\u03A6', 'Phi'),
('\u03A7', 'Chi'),
('\u03A8', 'Psi'),
('\u03A9', 'Omega'),
('\u03B1', 'alpha'),
('\u03B2', 'beta'),
('\u03B3', 'gamma'),
('\u03B4', 'delta'),
('\u03B5', 'epsilon'),
('\u03B6', 'zeta'),
('\u03B7', 'eta'),
('\u03B8', 'theta'),
('\u03B9', 'iota'),
('\u03BA', 'kappa'),
('\u03BB', 'lamda'),
('\u03BC', 'mu'),
('\u03BD', 'nu'),
('\u03BE', 'xi'),
('\u03BF', 'omicron'),
('\u03C0', 'pi'),
('\u03C1', 'rho'),
('\u03C3', 'sigma'),
('\u03C4', 'tau'),
('\u03C5', 'upsilon'),
('\u03C6', 'phi'),
('\u03C7', 'chi'),
('\u03C8', 'psi'),
('\u03C9', 'omega'),
)

_UNICODE_TO_GREEK = dict(_UNICODE_GREEK)
_GREEK_TO_UNICODE = dict([(x[1], x[0]) for x in _UNICODE_GREEK])

def unicode_to_greek(s):
    """Converts unicode single code, e.g., '\u03A3' to Greek letter name, e.g. 'Sigma'"""

    # "?" is the "zero-element"
    if s == "?":
        return s

    return _UNICODE_TO_GREEK[s]

def greek_to_unicode(s):
    """Converts Greek letter name, e.g., 'Sigma', to unicode character, e.g. '\u03A3' """

    # "?" is the "zero-element"
    if s == "?":
        return s

    return _GREEK_TO_UNICODE[s]



# superscript numbers
_INT_TO_SUPERSCRIPT = {
 0: "\u2070",
 1: "\u2071",
 2: "\u00b2",
 3: "\u00b3",
 4: "\u2074",
 5: "\u2075",
 6: "\u2076",
 7: "\u2077",
 8: "\u2078",
 9: "\u2079",
}

def int_to_superscript(i):
    """int_to_superscript(i) --> str"""

    return "".join((_INT_TO_SUPERSCRIPT[int(ch)] for ch in str(i)))













from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from . import WBase


__all__ = ["WDBRegistry"]


class WDBRegistry(WBase):
    """
    Base class for registry widgets

    Comes with
      - table widget
      - tool bar
      - set of actions
      - signals: id_changed, changed
      - 'id' and 'row' properties
    """

    # Emitted whenever the value of the 'id' property changes
    id_changed = pyqtSignal()

    @property
    def id(self):
        """Molecule id or None"""
        return self._get_id()

    @property
    def row(self):
        """Returns current data row: MyDBRow object, or None"""
        ret = None
        i = self.tableWidget.currentRow()
        if i >= 0:
            ret = self._data[i]
        return ret

    @property
    def f(self):
        """Object representing the file being edited (possibly a DataFile object)"""
        return self._f

    @f.setter
    def f(self, f):
        self._f = f
        self._f_changed()

    def __init__(self, *args):
        WBase.__init__(self, *args)

        # # Internal state
        self._flag_populating = False  # activated when populating table
        self._data = []
        self._last_id = None  # to control emitting id_changed signal
        # FileSQLiteDB instance or None
        self._f = None

        # # GUI design

        # ## Actions
        action = self.action_insert = QAction(a99.get_icon("list-add"), "&Insert...", self)
        action.triggered.connect(self.on_insert)
        action.setShortcut("Ins")

        action = self.action_edit = QAction(a99.get_icon("gtk-edit"), "&Edit...", self)
        action.triggered.connect(self.on_edit)
        action.setShortcut("Enter")

        action = self.action_delete = QAction(a99.get_icon("trash"), "&Delete", self)
        action.setShortcut("Del")
        action.triggered.connect(self.on_delete)

        # ## Visual design

        l = QHBoxLayout(self)
        a99.set_margin(l, 2)
        l.setSpacing(3)

        a = self.tableWidget = QTableWidget()
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(a99.MONO_FONT)
        a.setAlternatingRowColors(True)
        # TODO a.setSortingEnabled(True)
        a.installEventFilter(self)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)
        a.setContextMenuPolicy(Qt.CustomContextMenu)
        a.customContextMenuRequested.connect(self.on_tableWidget_customContextMenuRequested)
        a.itemDoubleClicked.connect(self.on_tableWidget_doubleClicked)

        l.addWidget(a)

        tb = QToolBar()
        tb.setOrientation(Qt.Vertical)
        l.addWidget(tb)
        tb.addAction(self.action_insert)
        tb.addAction(self.action_edit)
        tb.addAction(self.action_delete)


    def eventFilter(self, source, event):
        # TODO may have to use this if I insert the widget where there is more than one action with same Ins-Enter-Del short cuts (don't know)
        # if event.type() == QEvent.KeyPress:
        #     if event.key() == Qt.Key_Return:
        #         if source == self.tableWidget:
        #             self.on_edit()
        #             return True
        #             # if event.key() == Qt.Key_Delete:
        #             # ...
        return False


    # # "Abstract" methods (must be overriden)

    def _move_to_first(self):
        """Makes top left cell the current cell, if table is not empty"""
        t = self.tableWidget
        if t.rowCount() > 0:
            t.setCurrentCell(0, 0)

    def _do_on_insert(self):
        """Override this and return True if data was changed"""
        raise NotImplementedError()


    def _do_on_edit(self):
        """Override this and return True if data was changed"""
        raise NotImplementedError()


    def _do_on_delete(self):
        """Override this and return True if data was changed"""
        raise NotImplementedError()


    # # Slots

    def on_insert(self):
        if self._do_on_insert():
            self.changed.emit()

    def on_edit(self):
        if self.row is None:
            return
        if self._do_on_edit():
            self.changed.emit()

    def on_delete(self):
        if not self.row:
            return
        if self._do_on_delete():
            self.changed.emit()

    def on_tableWidget_customContextMenuRequested(self, position):
        menu = QMenu()
        menu.addAction(self.action_insert)
        menu.addAction(self.action_edit)
        menu.addAction(self.action_delete)
        action = menu.exec_(self.tableWidget.mapToGlobal(position))


    def on_tableWidget_doubleClicked(self):
        self.on_edit()


    def on_tableWidget_currentCellChanged(self, currentRow, currentColumn, previousRow,
                                          previousColumn):
        if not self._flag_populating:
            self._wanna_emit_id_changed()


    # # Internal gear

    # TODO work with sorting
    def _find_id(self, id_):
        """Moves to row where formula is (if found, otherwise does nothing)"""
        for i, row in enumerate(self._data):
            if row["id"] == id_:
                t = self.tableWidget
                # idx = t.itemFromIndex()
                t.setCurrentCell(i, 0)
                break

    def _wanna_emit_id_changed(self):
        """Filters intentions to emit the id_changed signal (only does if id really changed)"""
        if self._last_id != self._get_id():
            self._last_id = self._get_id()
            self.id_changed.emit()

    def _get_id(self):
        """Getter because using the id property from within was not working"""
        ret = None
        row = self.row
        if row:
            ret = row["id"]
        return ret


    def _f_changed(self):
        """Override this to react on whenever the value of the 'f' property changes"""


"""
Search tools
"""

__all__ = [
"index_nearest", "BSearch", "BSearchRound", "BSearchCeil", "BSearchFloor", "FindNotNaNBackwards", ]



from bisect import *
import numpy as np



def index_nearest(array, value):
    """
    Finds index of nearest value in array.

    http://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    """
    idx = (np.abs(array-value)).argmin()
    return idx



def BSearch(a, x, lo=0, hi=None):
    """Returns index of x in a, or -1 if x not in a.

    Arguments:
      a -- ordered numeric sequence
      x -- element to search within a
      lo -- lowest index to consider in search*
      hi -- highest index to consider in search*

    *bisect.bisect_left capability that we don't need to loose."""
    if len(a) == 0: return -1
    hi = hi if hi is not None else len(a)
    pos = bisect_left(a, x, lo, hi)
    return pos if pos != hi and a[pos] == x else -1  # doesn't walk off the end


def BSearchRound(a, x, lo=0, hi=None):
    """Returns index of a that is closest to x.

    Arguments:
      a -- ordered numeric sequence
      x -- element to search within a
      lo -- lowest index to consider in search*
      hi -- highest index to consider in search*

    *bisect.bisect_left capability that we don't need to loose."""
    if len(a) == 0: return -1
    hi = hi if hi is not None else len(a)
    pos = bisect_left(a, x, lo, hi)

    if pos >= hi:
        return hi - 1
    elif a[pos] == x or pos == lo:
        return pos
    else:
        return pos - 1 if x - a[pos - 1] <= a[pos] - x else pos


def BSearchCeil(a, x, lo=0, hi=None):
    """Returns lowest i such as a[i] >= x, or -1 if x > all elements in a

    So, if x is in between two elements in a, this function will return the
    index of the higher element, hence "Ceil".

    Arguments:
      a -- ordered numeric sequence
      x -- element to search within a
      lo -- lowest index to consider in search
      hi -- highest index to consider in search"""
    if len(a) == 0: return -1
    hi = hi if hi is not None else len(a)
    pos = bisect_left(a, x, lo, hi)
    return pos if pos < hi else -1


def BSearchFloor(a, x, lo=0, hi=None):
    """Returns highest i such as a[i] <= x, or -1 if x < all elements in a

    So, if x is in between two elements in a, this function will return the
    index of the lower element, hence "Floor".

    Arguments:
      a -- ordered numeric sequence
      x -- element to search within a
      lo -- lowest index to consider in search
      hi -- highest index to consider in search"""
    if len(a) == 0: return -1
    hi = hi if hi is not None else len(a)
    pos = bisect_left(a, x, lo, hi)
    return pos - 1 if pos >= hi \
        else (pos if x == a[pos] else (pos - 1 if pos > lo else -1))


def FindNotNaNBackwards(x, i):
    """Returns last position (starting at i backwards) which is not NaN, or -1."""
    while i >= 0:
        if not np.isnan(x[i]):
            return i
        i -= 1
    return -1



"""Routines that somehow look into the package itself"""


import os
import glob
import imp
import importlib
import inspect
import a99


__all__ = ["import_module", "ExeInfo", "get_exe_info", "collect_doc",
           "get_classes_in_module"]


def import_module(filename):
    """
    Returns module object

    Source: https://www.blog.pythonlibrary.org/2016/05/27/python-201-an-intro-to-importlib/
    """
    module_name = "xyz"

    module_spec = importlib.util.spec_from_file_location(module_name, filename)

    if module_spec is None:
        raise RuntimeError("Python cannot import file '{}'".format(filename))

    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    # print(dir(module))
    #
    # msg = 'The {module_name} module has the following methods:' \
    #       ' {methods}'
    # print(msg.format(module_name=module_name,
    #                  methods=dir(module)))

    return module


class ExeInfo(object):
    """Information about an executable file"""
    def __init__(self, filename, description, flag_error=False, flag_gui=False):
        self.filename = filename
        # First line of script docstring
        self.description = description
        # Error importing script as module?
        self.flag_error = flag_error
        # Script is a graphical application?
        self.flag_gui = flag_gui


def get_exe_info(dir_, flag_protected=False):
    """
    Returns a list of ExeInfo objects, which represent Python scripts within dir_

    Args:
        dir_: string, path to directory
        flag_protected: whether or not to include files starting with a '_'

    Returns:
        list of ExeInfo objects

    The ExeInfo objects represent the ".py" files in directory dir_,
    """

    ret = []
    # gets all scripts in script directory
    ff = glob.glob(os.path.join(dir_, "*.py"))
    # discards scripts whose file name starts with a "_"
    ff = [f for f in ff if flag_protected or not os.path.basename(f).startswith("_")]
    ff.sort()

    for f in ff:
        _, filename = os.path.split(f)
        flag_error = False
        flag_gui = None
        try:
            # Checks if it is a graphical application

            with open(f, "r") as h:
                flag_gui = "QApplication" in h.read()

            script_ = imp.load_source('script_', f)  # module object
            descr = script_.__doc__.strip()
            descr = descr.split("\n")[0]  # first line of docstring
        except Exception as e:
            flag_error = True
            descr = "*{0!s}*: {1!s}".format(e.__class__.__name__, str(e))

        ret.append(ExeInfo(filename, descr, flag_error, flag_gui))

    # Sorts command-line and graphical applications by name separately
    sisi_gra = [si for si in ret if si.flag_gui]
    sisi_cmd = [si for si in ret if not si.flag_gui]
    sisi_gra = sorted(sisi_gra, key=lambda x: x.filename)
    sisi_cmd = sorted(sisi_cmd, key=lambda x: x.filename)
    ret = sisi_cmd+sisi_gra

    return ret


def collect_doc(module, base_class=None, prefix="", flag_exclude_prefix=False):
    """
    Collects class names and docstrings in module for classes starting with prefix

    Arguments:
        module -- Python module
        prefix -- argument for str.startswith(); if not passed, does not filter
        base_class -- filters only descendants of this class
        flag_exclude_prefix -- whether or not to exclude prefix from class name in result

    Returns: [(classname0, signature, docstring0), ...]
    """

    ret = []
    for attrname in module.__all__:
        if prefix and not attrname.startswith(prefix):
            continue

        attr = module.__getattribute__(attrname)

        if base_class is not None and not issubclass(attr, base_class):
            continue

        spec = inspect.signature(attr)

        ret.append((attrname if not flag_exclude_prefix else attrname[len(prefix):], spec, attr.__doc__))

    return ret


def get_classes_in_module(module, superclass=object):
    """
    Returns a list with all classes in module that descend from parent

    Args:
        module: builtins.module
        superclass: a class

    Returns: list
    """

    ret = []
    for classname in dir(module):
        attr = module.__getattribute__(classname)
        try:
            if issubclass(attr, superclass) and (attr != superclass):
                ret.append(attr)
        except TypeError:
            # "issubclass() arg 1 must be a class"
            pass
        except RuntimeError:
            # a99.get_python_logger().exception("Failed probing attribute '{}'".format(classname))
            # raise
            pass
    return ret


def get_class_package(class_):
    """
    Returns package that class belongs to
    """
    root_pkg_name = class_.__module__.split(".")[0]
    if root_pkg_name == a99.__name__:
        return a99
    if root_pkg_name not in __collaborators:
        raise RuntimeError("Class '{}' belongs to package '{}', "
                           "but the latter is not among hypydrive collaborators".
                           format(class_.__name__, root_pkg_name))
    return __collaborators[root_pkg_name]




__all__ = ["Parameter", "Parameters"]


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from collections import OrderedDict


class Parameters(object):
    """
    Collection of parameters (for GUI purposes only)

    Arguments:
      specs -- [(name, {...}), ...] see Parameter.FromSpec() for full documentation.
    """
    def __init__(self, specs=[]):
        # List of Parameter objects
        self.params = []
        self._FromSpecs(specs)

    def __len__(self):
        return len(self.params)

    def __iter__(self):
        return self.params.__iter__()

    def _FromSpecs(self, specs):
        """
        Populates _params using specification

        Arguments:
          specs -- either:
            (a) list as [(name, {...}), ...] (see Parameter.FromSpec() for further information)
            (b) dictionary as {"name": value, ...}
        """

        if isinstance(specs, dict):
            specs_ = []
            for name, value in specs.items():
                specs_.append((name, {"value": value}))
        else:
            specs_ = specs
        for spec in specs_:
            self.params.append(Parameter(spec))

    def AddToLayout(self, layout):
        """
        Arguments:
          layout -- a QFormLayout instance
        """
        for param in self.params:
            widget = param.RenderWidget()
            layout.addRow(param.caption, widget)

    def UpdateFromWidgets(self):
        for param in self.params:
            try:
                param.UpdateValueFromWidget()
            except Exception as e:
                raise
                # raise type(e)("Error in '%s': %s" % (param.name, +str(e)), None, sys.exc_info()[2]


    def GetKwargs(self):
        ret = OrderedDict()
        for param in self.params:
            ret[param.name] = param.value
        return ret


class Parameter(object):
    """Definition of a single parameter, with QWidget rendering facility"""
    def __init__(self, spec=None):
        self.name = None
        self.caption = None
        self.toolTip = None
        self.type = None
        self.value = None
        self.widget = None  # Widget that will edit the thing
        if spec is not None:
            self.FromSpec(spec)

    def FromSpec(self, spec):
        """
        Args:
          spec: (name, {...}), or Parameter object

        Dict keys:
          "caption" -- (optional) text for label in editor. Defaults to the
                 keyword argument name
          "toolTip" (optional)
          "type" -- (optional, defaults to type("value") or int if "value" is
              not specified. Accepts:
                - int
                - float
                - str
                - bool
                - list
          "value" -- (optional) defaults to 1 if numeric, False if bool,
               "" if str

        """
        if isinstance(spec, Parameter):
            self.name = spec.name
            self.caption = spec.caption if spec.caption is not None else spec.name
            self.toolTip = spec.toolTip if spec.toolTip is not None else ""
            self.type = spec.type if spec.type is not None else type(spec.value) if spec.value is not None else int
            self.value = spec.value
        else:
            self.name, d = spec
            self.caption = d.get("caption", self.name)
            self.toolTip = d.get("toolTip", "")
            t = self.type = d.get("type", type(d["value"]) if "value" in d else int)
            if not t in (int, float, bool, str, list):
                raise TypeError("Invalid type: '{0!s}'".format(t.__name__))
            self.value = d.get("value")

        if self.value is None:
            self.value = 0 if self.type == int else \
                0. if self.type == float else \
                False if self.type == bool else ""

    def RenderWidget(self):
        """Returns a QWidget subclass instance. Exact class depends on self.type"""
        t = self.type
        if t == int:
            ret = QSpinBox()
            ret.setMaximum(999999999)
            ret.setValue(self.value)
        elif t == float:
            ret = QLineEdit()
            ret.setText(str(self.value))
        elif t == bool:
            ret = QCheckBox()
            ret.setChecked(self.value)
        else:  # str, list left
            ret = QLineEdit()
            ret.setText(str(self.value))
        self.widget = ret
        return ret

    def UpdateValueFromWidget(self):
        t, w = self.type, self.widget
        if t == int:
            self.value = int(w.value())
        elif t == float:
            self.value = float(eval(str(w.text())))
        elif t == bool:
            self.value = w.isChecked()
        elif t == str:
            self.value = str(w.text())
        else:  # list
            _value = eval(str(w.text()))
            if not isinstance(_value, list):
                raise ValueError("Parameter '{}' must evaluate to a list".format(self.name))
            self.value = _value



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '.\XParametersEditor.ui'
#
# Created: Fri Jul 18 20:44:20 2014
#    by: PyQt5 UI code generator 4.10.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99


try:
  _encoding = QApplication.UnicodeUTF8
  def _translate(context, text, disambig):
    return QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
  def _translate(context, text, disambig):
    return QApplication.translate(context, text, disambig)

class Ui_XParametersEditor(object):
  def setupUi(self, XParametersEditor):
    XParametersEditor.setObjectName("XParametersEditor")
    XParametersEditor.setWindowModality(Qt.WindowModal)
    XParametersEditor.resize(446, 62)
    self.horizontalLayout_2 = QHBoxLayout(XParametersEditor)
    self.horizontalLayout_2.setSpacing(2)
    a99.set_margin(self.horizontalLayout_2, 2)
    self.horizontalLayout_2.setObjectName("horizontalLayout_2")
    self.horizontalLayout = QHBoxLayout()
    self.horizontalLayout.setSpacing(2)
    self.horizontalLayout.setObjectName("horizontalLayout")
    self.frame = QFrame(XParametersEditor)
    self.frame.setFrameShape(QFrame.StyledPanel)
    self.frame.setFrameShadow(QFrame.Raised)
    self.frame.setObjectName("frame")
    self.horizontalLayout.addWidget(self.frame)
    self.buttonBox = QDialogButtonBox(XParametersEditor)
    self.buttonBox.setOrientation(Qt.Vertical)
    self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
    self.buttonBox.setObjectName("buttonBox")
    self.horizontalLayout.addWidget(self.buttonBox)
    self.horizontalLayout_2.addLayout(self.horizontalLayout)

    self.retranslateUi(XParametersEditor)
    self.buttonBox.rejected.connect(XParametersEditor.reject)
    self.buttonBox.accepted.connect(XParametersEditor.accept)
    QMetaObject.connectSlotsByName(XParametersEditor)

  def retranslateUi(self, XParametersEditor):
    XParametersEditor.setWindowTitle(_translate("XParametersEditor", "Parameters Editor", None))




from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
from .a_XLogDialog import XLogDialog
from .a_XLogMainWindow import XLogMainWindow
from .a_WBase import *
from . import xmisc
from .. import loggingaux
import os

# Margin for main layout
EDITOR_LAYMN_MAIN = 9
# Spacing for main layout
EDITOR_LAYSP_MAIN = 9

__all__ = ["WEditor", "EDITOR_LAYMN_MAIN", "EDITOR_LAYSP_MAIN"]


class WEditor(WBase):
    """Base class for editor widgets"""

    @property
    def f(self):
        # DataFile instance
        return self._f

    @property
    def flag_valid(self):
        return self._flag_valid

    # Emitted by load()
    loaded = pyqtSignal()

    def __init__(self, *args, **kwargs):
        WBase.__init__(self, *args, **kwargs)

        # Datafile instance
        self._f = None
        # whether file can be saved or not. Handling this is almost completely left to descenadnt
        self._flag_valid = False

        self.setEnabled(False)

        # # Central layout
        l = self.layout_editor = QVBoxLayout()
        xmisc.set_margin(l, EDITOR_LAYMN_MAIN)
        l.setSpacing(EDITOR_LAYSP_MAIN)
        self.setLayout(l)


        ll = self.layout_label_fn = QHBoxLayout()
        l.addLayout(ll)
        xmisc.set_margin(ll, 0)
        ll.setSpacing(3)

        la = self.label_caption_fn = QLabel("<b>File: </b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed))
        ll.addWidget(la)

        la = self.label_fn = QLabel()  # Label that shows the filename
        la.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed))
        ll.addWidget(la)

        ll.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

    def load(self, fobj):
        flag_ok = False
        try:
            self._do_load(fobj)

            # We trust subclass on this
            self._flag_valid = True

            self.update_gui_label_fn()
            self.setEnabled(True)
            flag_ok = True
        except Exception as E:
            loggingaux.get_python_logger().exception("Loading obj of class '{}'".format(fobj.__class__.__name__))
            xmisc.show_error(loggingaux.str_exc(E))
            self._flag_valid = False
            self._f = None

            # **Note** there is one issue: if fails here, it may leave the GUI partially updated

        if flag_ok:
            self.loaded.emit()

    def update_gui_label_fn(self):
        if hasattr(self, "label_fn"):
            self.label_fn.setText(self._make_fn_text())

    def _make_fn_text(self):
        """Makes filename text"""
        if not self._f:
            text = "(not loaded)"
        elif self._f.filename:
            text = os.path.relpath(self._f.filename, ".")
        else:
            text = "(filename not set)"
        return text

    def _do_load(self, fobj):
        raise NotImplementedError("Please implement {}._do_load()".format(self.__class__.__name__))


import os
import a99


__all__ = ["ErrorCollector", "Occurrence"]


class Occurrence(object):
    """Class that carries information about an error/warning, with HTML rendering ability"""
    colors = {"undefined": "#000000",
              "warning": a99.COLOR_WARNING,
              "error": a99.COLOR_ERROR,
              "cannot open": a99.COLOR_ERROR}

    def __init__(self, filename, type_, line, message):
        # "undefined"/"warning"/"error"/"cannot open"
        self.type = type_
        self.message = message
        self.filename = filename
        self.line = line

    def get_html(self):
        message = self.message if isinstance(self.message, str) \
         else "\n".join(self.message)
        return ("<h3>File '{0!s}' (<span style=\"color: {1!s}\">{2!s}</span>)</h3>\n".format(self.filename, self.colors[self.type], self.type))+ \
         ("" if not self.line else "<b>line {0:d}</b>\n".format(self.line))+ \
         ("<pre>{0!s}</pre>".format(message) if message else "")


class ErrorCollector(object):
    """Walks through directory in search for files 'python.log' and 'fortran.log'.

    Opens these files and extracts error information from them (if found).
    """

    def __init__(self):
        # List of Occurrence
        self.occurrences = []

    def collect_errors(self, path_):
        self.occurrences = []
        for p, dirs, files in os.walk(path_):
            for file in files:
                if file == "python.log":
                    pass
                    # parses python log
                elif file == "fortran.log":
                    path__ = os.path.relpath(os.path.join(p, file), ".") # os.path.relpath(os.path.join(path_, file), path_)
                    try:
                        linetext_last = ""
                        line, flag_halting = 1, False
                        with open(path__, "r") as h:
                            for t in h:
                                linetext = t.strip()
                                if "HALTING" in linetext:
                                    if not flag_halting:
                                        occ = Occurrence(path__, "error", line, [])
                                        self.occurrences.append(occ)
                                        flag_halting = True
                                    occ.message.append(linetext)
                                else:
                                    flag_halting = False

                                if "Fortran runtime error" in linetext:
                                    occ = Occurrence(path__, "error", line,
                                     [linetext_last, linetext])
                                    self.occurrences.append(occ)

                                if "WARNING" in linetext:
                                    occ = Occurrence(path__, "warning", line, linetext)
                                    self.occurrences.append(occ)

                                line += 1
                                linetext_last = linetext
                    except IOError as E:
                        occ = Occurrence(path__, "cannot open", 0, str(E))
                        self.occurrences.append(occ)

    def get_html(self):
        oo = [occ.get_html() for occ in self.occurrences]
        if len(oo) == 0:
            return "No errors were found."
        return "\n".join(oo)




"""Parameters Editor Widget. Mostly used embedded in XParametersEditor"""

__all__ = ["WParametersEditor"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .parameter import Parameters

################################################################################

class WParametersEditor(QWidget):
    """
    Parameters editor widget.

    Arguments:
      parent=None
      specs=None -- list as [(name, {...}), ...], which will be passed
       to Parameters constructor. Parameter.FromSpec()
       for full documentation.
      parameters=None -- Parameters instance. If passed, ignores specs and sets
      internal parameters attribute directly
    """

    # Emitted whenever the parameters changes into a valid parameters
    # TODO Not sure if will be used
    ParametersChanged = pyqtSignal(tuple)

    def __init__(self, parent=None, specs=None, parameters=None):
        QWidget.__init__(self, parent)

        if parameters is not None:
            assert isinstance(parameters, Parameters)
            self._parameters = parameters
        else:
            self._parameters = Parameters(specs)

        self._SetupUi()

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.formLayout.itemAt(0, QFormLayout.FieldRole).widget().setFocus()

    def get_kwargs(self):
        self._parameters.UpdateFromWidgets()
        return self._parameters.GetKwargs()

    def validate(self):
        # Currently calling Parameters.GetKwargs() to do the <quote>validation<quote>
        self._parameters.UpdateFromWidgets()

    def set_error_text(self, x):
        self._set_error_text(x)


    ############################################################################
    ############################################################################

    def _set_error_text(self, x):
        """Sets text of labelError."""
        self.labelError.setText(x)

    def _SetupUi(self):
        self.setObjectName("WParametersEditor")
        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.setSpacing(0)

        self.formLayout = QFormLayout()
        self.formLayout.setVerticalSpacing(4)
        self.formLayout.setHorizontalSpacing(5)
        self.verticalLayout.addLayout(self.formLayout)

        self.labelError = QLabel(self)
        self.labelError.setStyleSheet("color: red")
        self.labelError.setText("")
        self.verticalLayout.addWidget(self.labelError)

        spacerItem = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)

        self._parameters.AddToLayout(self.formLayout)



# # Temporary imports
#   =================
# These modules should be be del'eted at the end
import sys
import logging


# # Prevents from running in Python 2
#   =================================
# (http://stackoverflow.com/questions/9079036)
if sys.version_info[0] < 3:
    raise RuntimeError("Python version detected:\n*****\n{0!s}\n*****\nCannot run, must be using Python 3".format(sys.version))

# Antecipated import because of str_exc() used below
from .loggingaux import *


# # Initializes matplotlib to work with Qt4
#   =======================================
# Tk backend problems:
# - plot windows pop as modal
# - configuration options not as rich as Qt4Agg
# However, if another backend has been chosen before importing astroapi, will not choose Qt4 backend
flag_matplotlib = True
try:
    def init_agg():
        import matplotlib
        matplotlib.use('Qt5Agg')
    if not 'matplotlib.backends' in sys.modules:
        init_agg()
except ImportError as e:
    flag_matplotlib = False
    logging.warning("Error importing matplotlib: '{}', "
        "matplotlib-related resources will not available".format(str_exc(e)))

# # Setup
#   =====

# ## Constants affecting the Graphical User Interface style
#    ------------------------------------------------------

# ### Color definition
# Error color
COLOR_ERROR = "#AA0000" # sortta wine
# Warning color
COLOR_WARNING = "#C98A00" # sortta yellow
# Default color for label text
COLOR_DESCR = "#222222"


# ## Constants affecting logging
#    ---------------------------
#
# If the following need change, this should be done before calling get_python_logger() for the
# first time

# Set this to make the python logger to log to the console. Note: will have no
# effect if changed after the first call to get_python_logger()
flag_log_console = True

# Set this to make the python logger to log to a file named "python.log".
# Note: will have no effect if changed after the first call to get_python_logger()
flag_log_file = False

# Logging level for the python logger
logging_level = logging.INFO


from .conversion import *
from .debugging import *
from .fileio import *
from .introspection import *
if flag_matplotlib:
    from .matplotlibaux import *
from .misc import *
from .parts import *
from .search import *
from .textinterface import *
from .config import *
from .litedb import *
from .gui import *
from .datetimefunc import *
try:
    from .astropyaux import *
except ImportError as e:
    logging.warning("Error importing astropy: '{}', "
                    "astropy-related resources will not available".format(str_exc(e)))

# from . import conversion
# from . import debugging
# from . import fileio
# from . import introspection
# from . import loggingaux
# from . import matplotlibaux
# from . import misc
# from . import parts
# from . import search
# from . import textinterface
# from . import config
from . import gui

# # Finally, gets rid of unwanted symbols in the workspace
#   ======================================================
del init_agg, sys, logging  # Don't want this in the namespace



"""Parameters Editor dialog."""

__all__ = ["XParametersEditor"]

from .ui_XParametersEditor import Ui_XParametersEditor
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from . import a_WParametersEditor


################################################################################
class XParametersEditor(QDialog, Ui_XParametersEditor):
  """
  Arguments:
    parent=None -- nevermind
    specs=None -- *Required*! list as [(name, {...}), ...], which will be passed
     to Parameters constructor. Parameter.FromSpec()
     for full documentation.
  """

  def __init__(self, parent=None, specs=None, title=None):
    ## State variables
    QDialog.__init__(self, parent)
    self._ReadConfig()
    self.setupUi(self)

    pe = self.pe = a_WParametersEditor.WParametersEditor(specs=specs)
    self.wsbLayout = QVBoxLayout(self.frame)
    self.wsbLayout.setContentsMargins(0, 0, 0, 0)
    self.frame.layout().addWidget(pe)
    pe.setFocus()

    if title is not None:
      self.setWindowTitle(title)


  def get_kwargs(self):
    return self.pe.get_kwargs()

  def _ReadConfig(self):
    pass

  def accept(self):
    try:
      self.pe.validate()
    except Exception as E:
      QMessageBox.critical(None, "Error", str(E))

    QDialog.accept(self)

  def show(self):
    super(XParametersEditor, self).show()





"""
Gear for working with configuration files

  - subclass of `configobj.ConfigObj`
"""

from configobj import ConfigObj
import a99
import os
import re

__all__ = ["AAConfigObj", "get_config_obj"]

class AAConfigObj(ConfigObj):
    """Subclassed ConfigObj to create section structures automatically when getting/setting items"""

    def _get_section(self, path_):
        """Auto-creates section structure

        Last element in path_ is considered to be the "filename" (item name)

        Returns: (configobj.Session object, converted path)
        """

        if isinstance(path_, str):
            path_ = path_.strip().split("/")
            if path_[0] == "":
                path_ = path_[1:]

        obj = self
        for section_name in path_[:-1]:
            try:
                obj = obj[section_name]
            except KeyError:
                obj[section_name] = {}  # creates section
                obj = obj[section_name]

        return obj, path_

    def get_item(self, path_, default):
        """Returns item or default

        Arguments:
            path_ -- path to item in section/subsection structure. May be either:
                     ["section", "subsection", ...] or
                     "[/]section/subsection/..."  (leading slash is tolerated)
            default -- value to return if item is not found

        Argument 'default' is also used to determine the type of the data to return:

            - str and list: returned as retrieved
            - int and float: eval'ed
            - bool: parsed
        """
        section, path_ = self._get_section(path_)


        key = path_[-1]

        if key not in section:
            return default

        xvalue = section[key]
        type_ = type(default)
        if type_ in (str, list):
            return xvalue
        elif type_ == bool:
            value = True if xvalue == "True" else False if xvalue == "False" else eval(xvalue)
        elif type_ in (int, float):
            value = type_(xvalue)
        else:
            raise TypeError("Type not supported: {}".format(type_.__name__))
        return value


    def set_item(self, path_, value):
        """Sets item and automatically saves file"""
        section, path_ = self._get_section(path_)
        section[path_[-1]] = value
        self.write()


def get_config_obj(filename):
    """Reads/creates filename at user **home** folder and returns a AAConfigObj object"""

    if not filename.startswith("."):
        a99.get_python_logger().warning("Configuration filename '{}' does not start with a '.'".format(filename))

    path_ = os.path.join(os.path.expanduser("~"), filename)
    return AAConfigObj(path_, encoding="UTF8")



"""Matplotlib-related routines"""


__all__ = ["set_facecolor_white", "format_BLB", "format_legend", "set_figure_size", "remove_line"]


from matplotlib import rc, pyplot as plt
from matplotlib.ticker import AutoMinorLocator


def set_facecolor_white():
    rc("figure", facecolor="white")

set_facecolor_white()

def format_BLB():
    """Sets some formatting options in Matplotlib."""
    rc("figure", facecolor="white")
    rc('font', family = 'serif', size=10) #, serif = 'cmr10')
    rc('xtick', labelsize=10)
    rc('ytick', labelsize=10)
    rc('axes', linewidth=1)
    rc('xtick.major', size=4, width=1)
    rc('xtick.minor', size=2, width=1)
    rc('ytick.major', size=4, width=1)
    rc('ytick.minor', size=2, width=1)
    # plt.minorticks_on()
    #minorLocator = AutoMinorLocator()
    # plt.tick_params(which="both", width=2)
    # plt.tick_params(which="major", width=2)
    #rc('text', usetex=True)

def format_legend(leg):
    """Sets some formatting options in a matplotlib legend object."""
    # rect = leg.get_frame()
    # rect.set_linewidth(2.)

def set_figure_size(fig, width, height):
    """Sets MatPlotLib figure width and height in pixels

    Reference: https://github.com/matplotlib/matplotlib/issues/2305/
    """
    dpi = float(fig.get_dpi())
    fig.set_size_inches(float(width) / dpi, float(height) / dpi)



def remove_line(line2D):
    """
    Removes line from matplotlib plot.
    # http://stackoverflow.com/questions/4981815/how-to-remove-lines-in-a-matplotlib-plot
    """
    l = line2D.pop(0)
    l.remove()
    del l



"""
API to facilitate building a GUI

  - Reusable and subclassable widgets and windows
  - miscellanea of GUI-related features in `xmisc.py`
  - Python syntax highlighter
  -

File naming conventions:
  a_X*.py : classes descending from QMainWindow or QMainDialog
  a_W*.py : widgets

"""

import a99
import logging

# Tests PyQt5: if fails, 'gui' subpackage will be empty
flag_ok = True
try:
    from PyQt5.QtGui import QFont
except ImportError as e:
    logging.warning("Error importing PyQt5.QtCore: '{}', "
        "a99 GUI resources will not be available".format(a99.str_exc(e)))
    flag_ok = False

if flag_ok:
    # ### Standard font to be used in all GUIs
    MONO_FONT = QFont("not_a_font_name")
    MONO_FONT.setStyleHint(QFont.TypeWriter)

    del QFont

    from . import xmisc
    from . import parameter
    from . import syntax
    # The other modules (a_*) have too ugly names to be imported

    from .xmisc import *
    from .parameter import *
    from .syntax import *
    from .errorcollector import *
    from .a_WBase import *
    from .a_WParametersEditor import *
    from .a_WCollapsiblePanel import *
    from .a_XLogDialog import *
    from .a_XLogMainWindow import *
    from .a_XParametersEditor import *
    from .a_WDBRegistry import *
    from .a_WSelectFileOrDir import *
    from .a_WEditor import *
    from .a_WConfigEditor import *

del a99, flag_ok, logging



import os
from astropy.io import fits
import a99


__all__ = ["overwrite_fits"]


def overwrite_fits(hdulist, filename):
    """
    Saves a FITS file. Combined file rename, save new, delete renamed for FITS files
    Why: HDUlist.writeto() does not overwrite existing files
    Why(2): It is also a standardized way to save FITS files
    """

    assert isinstance(hdulist, (fits.HDUList, fits.PrimaryHDU))
    temp_name = None
    flag_delete_temp = False
    if os.path.isfile(filename):
        # PyFITS does not overwrite file
        temp_name = a99.rename_to_temp(filename)
    try:
        hdulist.writeto(filename, output_verify='warn')
        flag_delete_temp = temp_name is not None
    except:
        # Writing failed, reverts renaming
        os.rename(temp_name, filename)
        raise

    if flag_delete_temp:
        os.unlink(temp_name)





from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from .a_XParametersEditor import *
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT # as NavigationToolbar2QT
import matplotlib.pyplot as plt
import collections
import a99
from threading import Lock
import time
import os


__all__ = [
    "keep_ref",
    "check_return_space", "show_error", "enc_name_descr", "SignalProxy", "are_you_sure",
    "show_warning", "_window_titles", "style_checkboxes", "VerticalLabel", "show_message",
    "get_matplotlib_layout", "nerdify", "place_center", "get_window_title", "style_widget_changed",
    "get_QApplication", "get_icon", "show_edit_form", "snap_left", "snap_right", "enc_name",
    "place_left_top", "reset_table_widget", "table_info_to_parameters", "format_title0",
    "format_title1", "format_title2", "style_widget_valid", "set_margin",
    "get_frame", "set_checkbox_value"
]

_qrefs = []


def keep_ref(obj):
    """Keeps a reference to Python object representing a C++ Qt object

    **Rationale**: Qt windows may simply disappear when the Python object (e.g., QMainWindow())
                   representing a Qt object runs out of scope (i.e., the program leaves the method
                   that created the Python object). Python deletes the object and the C++ object
                   may be deleted as well. Keeping these objects in a list prevents this
                   from happening.

    **Drawback** will keep reference until the Python program finishes.

    Returns: obj
    """
    global _qrefs
    _qrefs.append(obj)
    return obj


def enc_name_descr(name, descr, color=a99.COLOR_DESCR):
    """Encodes html given name and description."""
    return enc_name(name, color)+"<br>"+descr


def enc_name(name, color=a99.COLOR_DESCR):
    """Encodes html given name."""
    return "<span style=\"color: {0!s}; font-weight: bold\">{1!s}</span>".format(color, name)


def style_checkboxes(widget):
    """
    Iterates over widget children to change checkboxes stylesheet.

    The default rendering of checkboxes does not allow to tell a focused one
    from an unfocused one.
    """

    ww = widget.findChildren(QCheckBox)
    for w in ww:
        w.setStyleSheet("QCheckBox:focus {border: 1px solid #000000;}")


def style_widget_changed(spinbox, flag_changed):
    """(Paints background yellow)/(removes stylesheet)"""
    spinbox.setStyleSheet("QWidget {background-color: #FFFF00}" if flag_changed else "")


def style_widget_valid(spinbox, flag_valid):
    """(Paints background pastel red)/(removes stylesheet)

    Reference: http://www.colorhexa.com/ff6961
    """
    spinbox.setStyleSheet("QWidget {background-color: #FF6961}" if not flag_valid else "")


def check_return_space(event, callable_):
    """Checks if event corresponds to Return/Space being pressed and calls callable_ if so."""
    if event.type() == QEvent.KeyPress:
        if event.key() in [Qt.Key_Return, Qt.Key_Space]:
            callable_()
            return True
    return False


def show_error(s):
  QMessageBox.critical(None, "Error", s)


def show_message(s):
  QMessageBox.information(None, "Information", s)


def show_warning(s):
  QMessageBox.warning(None, "Warning", s)


def are_you_sure(flag_changed, evt, parent=None, title="File has been changed",
                 msg="Are you sure you want to exit?"):
    """
    "Are you sure you want to exit" question dialog.

    If flag_changed, shows question dialog. If answer is not yes, calls evt.ignore()

    Arguments:
      flag_changed
      evt -- QCloseEvent instance
      parent=None -- parent form, used to centralize the question dialog at
      title -- title for question dialog
      msg -- text of question dialog

    Returns True or False. True means: "yes, I want to exit"
    """
    if flag_changed:
        r = QMessageBox.question(parent, title, msg,
             QMessageBox.Yes|QMessageBox.No, QMessageBox.Yes)
        if r != QMessageBox.Yes:
            evt.ignore()


def reset_table_widget(t, rowCount, colCount):
    """Clears and resizes a table widget."""
    t.reset()
    t.horizontalHeader().reset()
    t.clear()
    t.sortItems(-1)
    t.setRowCount(rowCount)
    t.setColumnCount(colCount)


def show_edit_form(obj, attrs, title):
    """Shows parameters editor modal form.

    Arguments:
      obj -- object to extract attribute values from
      attrs -- list of attribute names
    """
    specs = []
    for name in attrs:
        value = obj.__getattribute__(name)
        if value is None:
            value = ""  # None becomes str
        specs.append((name, {"value": value}))
    form = XParametersEditor(specs=specs, title=title)
    r = form.exec_()
    return r, form


# # Coarse solution for hidden window title
# Qt does not account for the window frame.This is being coarsely
# accounted for by setting the position coordinates to values slightly greater
# than 0.
_DESKTOP_OFFSET_LEFT = 2
_DESKTOP_OFFSET_TOP = 15


def place_left_top(window, width=None, height=None):
    """Places window in top left corner of screen.

    Arguments:
      window -- a QWidget
      width=None -- window width, in case you want to change it (if not passed, not changed)
      height=None -- window height, in case you want to change it (if not passed, not changed)
    """

    if width is None:
        width = window.width()
    if height is None:
        height = window.height()

    window.setGeometry(_DESKTOP_OFFSET_LEFT, _DESKTOP_OFFSET_TOP, width, height)


def place_center(window, width=None, height=None):
    """Places window in the center of the screen."""
    screenGeometry = QApplication.desktop().screenGeometry()

    w, h = window.width(), window.height()

    if width is not None or height is not None:
        w = width if width is not None else w
        h = height if height is not None else h
        window.setGeometry(0, 0, w, h)

    x = (screenGeometry.width() - w) / 2
    y = (screenGeometry.height() - h) / 2
    window.move(x, y)


def snap_left(window, width=None):
    """Snaps window to left of desktop.
    Arguments:
      window -- a QWidget
      width=None -- window width, in case you want to change it (if not passed, not changed)
    """
    if not width:
        width = window.width()
    rect = QApplication.desktop().screenGeometry()
    window.setGeometry(_DESKTOP_OFFSET_LEFT, _DESKTOP_OFFSET_TOP, width, rect.height())


def snap_right(window, width=None):
    """Snaps window to right of desktop.
    Arguments:
      window -- a QWidget
      width=None -- window width, in case you want to change it (if not passed, not changed)
    """
    if not width:
        width = window.width()
    rect = QApplication.desktop().screenGeometry()
    window.setGeometry(rect.width()-width, _DESKTOP_OFFSET_TOP, width, rect.height())


def nerdify(window):
    window.setFont(a99.MONO_FONT)


class VerticalLabel(QLabel):
    """Label that draws itself vertically.

    This was created to be used at lateral title:
      - It paints in bold
      - No HTML support
    """
    
    def paintEvent(self, evt):
        painter = QPainter(self)
        painter.setPen(Qt.black)
        painter.setBrush(Qt.Dense1Pattern)
        painter.rotate(90)
        painter.font().setWeight(QFont.Bold)

        # td = QTextDocument()
        # td.setHtml(self.text())
        # td.drawContents(painter)

        painter.drawText(0,0, self.text())

    def minimumSizeHint(self):
        s = QLabel.minimumSizeHint(self)
        return QSize(s.height(), s.width())

    def sizeHint(self):
        s = QLabel.sizeHint()
        return QSize(s.height(), s.width())


def get_matplotlib_layout(widget, flag_toolbar=True):
    """
    Creates figure, toolbar, layout, sets widget layout

    Returns figure, canvas, layout

    Reference: http://stackoverflow.com/questions/12459811

    """
    fig = plt.figure()
    canvas = FigureCanvas(fig)
    #        self.canvas.mpl_connect('button_press_event', self.on_plot_click)
    layout = QVBoxLayout(widget)
    if flag_toolbar:
        toolbar = NavigationToolbar2QT(canvas, widget)
        layout.addWidget(toolbar)
    layout.addWidget(canvas)
    a99.set_margin(layout, 0)

    return fig, canvas, layout


def get_icon(keyword):
    """
    Transforms a PNG file in a QIcon

    Looks for a file named <keyword>.png in the "icons" directory

    If file does not exist, returns None
    """

    filename = a99.get_path( "icons", keyword + ".png")
    if not os.path.isfile(filename):
        raise FileNotFoundError("File '{}' does not exist".format(filename))
    return QIcon(filename)


# Because several windows of the same class may be created, we'll give them different titles to help avoid confusion
_window_titles = collections.Counter()
def get_window_title(prefix):
    _window_titles[prefix] += 1
    i = _window_titles[prefix]
    if i == 1:
        return prefix
    else:
        return "{0!s} #{1:d}".format(prefix, i)


_qapp = None
def get_QApplication(args=[]):
    """Returns the QApplication instance, creating it is does not yet exist."""
    global _qapp
    if _qapp is None:
        QCoreApplication.setAttribute(Qt.AA_X11InitThreads)
        _qapp = QApplication(args)

    return _qapp


class _ThreadsafeTimer(QObject):
    """
    Thread-safe replacement for QTimer.

    Original author: Luke Campagnola -- pyqtgraph package
    """

    timeout = pyqtSignal()
    sigTimerStopRequested = pyqtSignal()
    sigTimerStartRequested = pyqtSignal(object)

    def __init__(self):
        QObject.__init__(self)
        self.timer = QTimer()
        self.timer.timeout.connect(self.timerFinished)
        self.timer.moveToThread(QCoreApplication.instance().thread())
        self.moveToThread(QCoreApplication.instance().thread())
        self.sigTimerStopRequested.connect(self.stop, Qt.QueuedConnection)
        self.sigTimerStartRequested.connect(self.start, Qt.QueuedConnection)


    def start(self, timeout):
        isGuiThread = QThread.currentThread() == QCoreApplication.instance().thread()
        if isGuiThread:
            #print "start timer", self, "from xgear thread"
            self.timer.start(timeout)
        else:
            #print "start timer", self, "from remote thread"
            self.sigTimerStartRequested.emit(timeout)

    def stop(self):
        isGuiThread = QThread.currentThread() == QCoreApplication.instance().thread()
        if isGuiThread:
            #print "stop timer", self, "from xgear thread"
            self.timer.stop()
        else:
            #print "stop timer", self, "from remote thread"
            self.sigTimerStopRequested.emit()

    def timerFinished(self):
        self.timeout.emit()


class SignalProxy(QObject):
    """Object which collects rapid-fire signals and condenses them
    into a single signal or a rate-limited stream of signals.
    Used, for example, to prevent a SpinBox from generating multiple
    signals when the mouse wheel is rolled over it.

    Emits sigDelayed after input signals have stopped for a certain period of time.

    Note: *queued* connection is made to slot.

    Original author: Luke Campagnola -- pyqtgraph package

    Arguments:
      signals -- a list of bound signals or pyqtSignal instance
      delay=0.3 -- Time (in seconds) to wait for signals to stop before emitting
      slot -- Optional function to connect sigDelayed to.
      rateLimit=0 -- (signals/second) if greater than 0, this allows signals to
       stream out at a steady rate while they are being received.
      flag_connect=True -- whether or not to start with the connections already
       made. If False, the signals and slots can be connected by calling
       connect_all()
    """

    __sigDelayed = pyqtSignal(object)

    def __init__(self, signals, delay=0.3, rateLimit=0, slot=None,
                 flag_connect=True):
        QObject.__init__(self)
        # for signal in signals:
        #     self.__connect_signal(signal)
        self.__signals = signals
        self.__delay = delay
        self.__rateLimit = rateLimit
        self.__args = None
        self.__timer = _ThreadsafeTimer()
        self.__timer.timeout.connect(self.__flush)
        self.__disconnecting = False
        self.__slot = slot
        self.__lastFlushTime = None
        self.__lock = Lock()
        # State: connected/disconnected
        self.__connected = False
        if flag_connect:
            self.connect_all()
        # if slot is not None:
        #     self.__sigDelayed.connect(slot, Qt.QueuedConnection)

    def add_signal(self, signal):
        """Adds "input" signal to connected signals.
        Internally connects the signal to a control slot."""
        self.__signals.append(signal)
        if self.__connected:
            # Connects signal if the current state is "connected"
            self.__connect_signal(signal)

    def connect_all(self):
        """[Re-]connects all signals and slots.

        If already in "connected" state, ignores the call.
        """
        if self.__connected:
            return  # assert not self.__connected, "connect_all() already in \"connected\" state"
        with self.__lock:
            for signal in self.__signals:
                self.__connect_signal(signal)
            if self.__slot is not None:
                self.__sigDelayed.connect(self.__slot, Qt.QueuedConnection)
            self.__connected = True

    def disconnect_all(self):
        """Disconnects all signals and slots.

        If already in "disconnected" state, ignores the call.
        """
        if not self.__connected:
            return  # assert self.__connected, "disconnect_all() already in \"disconnected\" state"
        self.__disconnecting = True
        try:
            for signal in self.__signals:
                signal.disconnect(self.__signalReceived)
            if self.__slot is not None:
                self.__sigDelayed.disconnect(self.__slot)
            self.__connected = False
        finally:
            self.__disconnecting = False

    def __signalReceived(self, *args):
        """Received signal. Cancel previous timer and store args to be forwarded later."""
        if self.__disconnecting:
            return
        with self.__lock:
            self.__args = args
            if self.__rateLimit == 0:
                self.__timer.stop()
                self.__timer.start((self.__delay * 1000) + 1)
            else:
                now = time.time()
                if self.__lastFlushTime is None:
                    leakTime = 0
                else:
                    lastFlush = self.__lastFlushTime
                    leakTime = max(0, (lastFlush + (1.0 / self.__rateLimit)) - now)

                self.__timer.stop()
                # Note: original was min() below.
                timeout = (max(leakTime, self.__delay) * 1000) + 1
                self.__timer.start(timeout)

    def __flush(self):
        """If there is a signal queued up, send it now."""
        if self.__args is None or self.__disconnecting:
            return False
        #self.emit(self.signal, *self.args)
        self.__sigDelayed.emit(self.__args)
        self.__args = None
        self.__timer.stop()
        self.__lastFlushTime = time.time()
        return True

    def __connect_signal(self, signal):
        signal.connect(self.__signalReceived, Qt.QueuedConnection)


def table_info_to_parameters(table_info):
    """
    Converts a list of MyDBRow into a parameters.Parameters object

    This facilitates transfering data from SQLite table row to a XParameterEditor window

    See also: get_table_info()
    """

    # Example of item in table_info:
    #   MyDBRow([('cid', 0), ('name', 'id'), ('type', 'integer'), ('notnull', 0), ('dflt_value', None), ('pk', 1)])

    opbj = a99.Parameters()
    for field_info in table_info.values():
        p = a99.Parameter()
        if field_info.type == "integer":
            p.type = int
        elif field_info.type == "real":
            p.type = float
        else:
            p.type = str

        p.name = field_info.name
        if field_info.dflt_value is not None:
            p.value = field_info.dflt_value
        opbj.params.append(p)
    return opbj


def format_title0(s):
    """Formats string as first-level title (to use as QLabel text)"""
    return "<h3>{}</h3>".format(s)


def format_title1(s):
    """Formats string as second-level title (to use as QLabel text)"""
    return "<h4>{}</h4>".format(s)


def format_title2(s):
    """Formats string as third-level title (to use as QLabel text)"""
    return "<h5>{}</h5>".format(s)


def set_margin(obj, margin):
    """...because Qt5 no longer has xxxx.setMargin() method"""
    obj.setContentsMargins(margin, margin, margin, margin)


def get_frame():
    """Returns a QFrame formatted in a particular way"""
    ret = QFrame()
    ret.setLineWidth(1)
    ret.setMidLineWidth(0)
    ret.setFrameShadow(QFrame.Sunken)
    ret.setFrameShape(QFrame.Box)
    return ret

def set_checkbox_value(w, value):
    """
    Sets a checkbox's "checked" property + signal blocking + value tolerance

    Args:
        w: QCheckBox instance
        value: something that can be converted to a bool
    """
    save = w.blockSignals(True)
    try:
        w.setChecked(bool(value))
    finally:
        w.blockSignals(save)




from PyQt5.QtWidgets import QDialog
from ._logpart import *

__all__ = ["XLogDialog"]

class XLogDialog(QDialog, _LogPart):
    """QDialog subclass that carries a logging facility"""
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        _LogPart.__init__(self)
        self._refs = []


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from . import WBase
import os

__all__ = ["WSelectDir", "WSelectFile"]


class _WSelectFileOrDir(WBase):
    @property
    def value(self):
        return self._get_value()

    @value.setter
    def value(self, value):
        self._set_value(value)

    @property
    def flag_valid(self):
        return self._flag_valid

    @flag_valid.setter
    def flag_valid(self, value):
        self._flag_valid = value
        self._update_gui()

    def __init__(self, *args, dialog_title="", dialog_path=".", dialog_wild="*"):
        a99.WBase.__init__(self, *args)

        self._last_value = None
        self._last_exists = None
        # This affects the edit background (red when _flag_valid is false
        self._flag_valid = True

        self._type = None
        self.dialog_title = dialog_title
        self.dialog_path = dialog_path
        self.dialog_wild = dialog_wild

        lw = QHBoxLayout()
        a99.set_margin(lw, 0)
        lw.setSpacing(4)
        self.setLayout(lw)

        # t = self.label = self.keep_ref(QLabel())
        # lw.addWidget(t)

        e = self.edit = QLineEdit()
        e.textEdited.connect(self._on_text_edited)
        lw.addWidget(e)
        # e.setReadOnly(True)

        b = self.button = QToolButton()
        b.clicked.connect(self._on_button_clicked)
        lw.addWidget(b)
        b.setFixedWidth(30)

        # Forces paint red if invalid
        self._on_text_edited()

    def _on_button_clicked(self, _):
        self._act_on_button_clicked()

    def _on_text_edited(self):
        self._act_on_change()

    def _update_gui(self):
        a99.style_widget_valid(self.edit, self._flag_valid)

    def _wanna_emit(self):
        value_now = self._get_value()
        # Maybe it is overzealous to only notify if the last valid filename is not the current one
        if True or value_now != self._last_value:
            self._last_value = value_now
            self.changed.emit()

    def _get_value(self):
        return self.edit.text().strip()

    def _set_value(self, value):
        if value is None:
            value = ""
        self.edit.setText(value)
        self._last_value = value
        self._last_exists = os.path.isfile(value)

    def _act_on_button_clicked(self):
        raise NotImplementedError()

    def _set_text_from_file_dialog(self, text):
        self.edit.setText(text)
        self.dialog_path = text
        self._act_on_change()

    def _act_on_change(self):
        exists = os.path.isfile(self._get_value())
        if not exists and self._last_exists:
            self._last_exists = False
            self.changed.emit()
        elif exists:
            self._last_exists = True
            self._wanna_emit()


class WSelectFile(_WSelectFileOrDir):
    def __init__(self, *args, **kwargs):
        _WSelectFileOrDir.__init__(self, *args, **kwargs)
        # self.label.setText("&File")
        self.button.setIcon(a99.get_icon("document-open"))

    def _on_button_clicked(self):
        path_ = self.edit.text().strip()
        if len(path_) == 0:
            path_ = self.dialog_path
        res = QFileDialog.getOpenFileName(self, self.dialog_title, path_,
                                          self.dialog_wild)[0]
        if res:
            self._set_text_from_file_dialog(res)


class WSelectDir(_WSelectFileOrDir):
    def __init__(self, *args, **kwargs):
        _WSelectFileOrDir.__init__(self, *args, **kwargs)
        self.button.setIcon(a99.get_icon("folder-change"))
        self.edit.setText(self.dialog_path)

    def _act_on_button_clicked(self):
        path_ = self.edit.text().strip()
        if len(path_) == 0:
            path_ = self.dialog_path
        res = QFileDialog.getExistingDirectory(self, self.dialog_title, path_)
        if len(res) > 0:
            self._set_text_from_file_dialog(res)


    def _validate(self):
        return os.path.isdir(self.value)



import sys
if sys.version_info[0] < 3:
    print("Python version detected:\n*****\n{0!s}\n*****\nCannot run, must be using Python 3".format(sys.version))
    sys.exit()

from setuptools import setup, find_packages
from glob import glob


def find_scripts(pkgs):
    ret = []
    for pkgname in pkgs:
        ret.extend(glob(pkgname + '/scripts/*.py'))
    return ret


pkgs = find_packages()
scripts = find_scripts(pkgs)


setup(
    name='a99',
    packages=find_packages(),
    include_package_data=True,
    version='0.17.10.26.0',
    license='GNU GPLv3',
    platforms='any',
    description='A multi-purpose API in Python',
    author='Julio Trevisan',
    author_email='juliotrevisan@gmail.com',
    url='https://github.com/trevisanj/a99',
    keywords= ['astronomy', 'pyqt', 'pyqt5', 'debugging', 'introspection', 'file', 'search', 'conversion', 'datetime', 'config', 'litedb', 'matplotlib', 'text'],
    install_requires=[],
    scripts=scripts
)


# TODO later install_requires=['numpy', 'matplotlib', 'pyqt5'],  # matplotlib never gets installed correctly by pip, but anyway...



from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
from .a_XLogDialog import XLogDialog
from .a_XLogMainWindow import XLogMainWindow
from .a_WEditor import *
from . import xmisc
import os
import traceback


# Margin for main layout
EDITOR_LAYMN_MAIN = 9
# Spacing for main layout
EDITOR_LAYSP_MAIN = 9

__all__ = ["WConfigEditor", "EDITOR_LAYMN_MAIN", "EDITOR_LAYSP_MAIN", "CEMapItem"]



# # https://stackoverflow.com/questions/3681272/can-i-get-a-reference-to-a-python-property
# def _get_dict_attr(obj, attr):
#     """I used this to get a property object from an object, not the property value"""
#     for obj in [obj] + obj.__class__.mro():
#         if attr in obj.__dict__:
#             return obj.__dict__[attr]
#     raise AttributeError
#
# # Short because it is going to be called many times
# GDA = _get_dict_attr()



class CEMapItem(object):
    """Config Editor map item

    There to way to get-set values. The order of precedence is as follows:

        (getter, setter) > (guiobj, propertyname)

    **Notes** on building an editor:
        - properties must block signals (implement self._after_update_gui() if necessary)
    """

    def __init__(self, fieldname, guiobj, propertyname=None, getter=None, setter=None):
        self.fieldname = fieldname
        # GUI object that has property called propertyname or getter-setter methods
        self.guiobj = guiobj
        self.propertyname = propertyname if propertyname is not None else fieldname
        self.getter = getter
        self.setter = setter

    def get_value(self):
        if self.getter:
            return self.getter()
        return self.guiobj.__getattribute__(self.propertyname)

    def set_value(self, value):
        if self.setter:
            self.setter(value)
        else:
            self.guiobj.__setattr__(self.propertyname, value)



class WConfigEditor(WEditor):
    """Base class for FileConfig editors"""

    def __init__(self, *args, **kwargs):
        WEditor.__init__(self, *args, **kwargs)
        
        # [(fieldname, read-write property object), ...]
        self._map = []
        
    ################################################################################################
    # Inheritable events
    
    def _before_update_gui(self):
        """Inherit this to tamper with self.f before it is "loaded" into the GUI"""
        
    def _before_update_fobj(self):
        """Inherit this method to perform validation before GUI state is "saved" into self.f"""

    def _after_update_fobj(self):
        """Inherit this to "tidy up" self.f after is is updated"""

    def _after_update_gui(self):
        """
        Inherit this carry out further GUI or internal state update sensible to self.fobj

        After self.f is "loaded" into the GUI by self.__do_update_gui(), there may be extra work
        left, such as extra calculations that take more than one config option into account.
        """

    ################################################################################################
    # Internal workings

    def _set_flag(self, w, value):
        """See a99.set_checkbox_value()"""
        xmisc.set_checkbox_value(w, value)

    def _do_load(self, fobj):
        self._f = fobj
        self._update_gui()
        self._flag_valid = True
        self.setEnabled(True)

    def _update_gui(self):
        "Updates GUI from fobj. Opposite of _update_fobj()"

        self._before_update_gui()
        self.__do_update_gui()
        self._after_update_gui()

    def __do_update_gui(self):
        for item in self._map:
            value = self._f.obj[item.fieldname]
            item.set_value(value)

    def _update_fobj(self):
        """Updates fobj from GUI. Opposite of _update_gui()."""

        # print("PPPPPPPPPPPPPPPPPPPRINTANDO O STACK")
        # traceback.print_stack()

        emsg, flag_error = "", False
        fieldname = None
        try:
            
            self._before_update_fobj()

            for item in self._map:
                fieldname = item.fieldname
                value = item.get_value()
                self._f.obj[fieldname] = value

            self._after_update_fobj()

        except Exception as E:
            flag_error = True
            if fieldname is not None:
                emsg = "Field '{}': {}".format(fieldname, str(E))
            else:
                emsg = str(E)
            self.add_log_error(emsg)

        self._flag_valid = not flag_error
        if not flag_error:
            self.status("")



"""Miscellanea of miscellanea"""


__all__ = ["random_name"]


import random


_forenames = ["Solomon", "John", "Loretta", "Stephen", "Harry", "Nancy", "Tracy", "Maggie", "Lafanda", "Napoleon", "Joe",
        "Ana", "Olivia", "Lucia", "Julien", "June", "Ada", "Blaise", "Platypus", "R2D2", "Obi-Wan",
        "Yoda", "Lancelot", "Shaun", "C3PO", "Luke", "George", "Martin", "Elvira", "Galileo", "Elizabeth",
        "Genie", "Mark", "Karl", "Henry-David", "Ludmilla", "Darth", "Bayden", "Plamen", "Margareth", "Javier",
        "Pouria", "Klen", "Lydiane", "Charlotte", "Edna", "Ricardo", "Francis", "Jemma", "Valon", "Imran", "Sian",
        "Hayat", "Taghreed", "Orla", "Michael", "Lourdes", "Weiyi", "Thomas", "Willian", "Miguel", "Rui",
        "Abdullah", "Angus", "Malcolm", "Donald", "Mickey", "Polona", "Rashmi", "Xiaowei", "Sasha", "Luciano",
        "Avinash", "Anthony", "Karen", "Matthew", "Tatiana", "Mariana", "Antonio", "Hamilton", "Pauderney",
        "BB-8"]
_surnames = ["Northupp", "Kanobi", "de Morgan", "de Vries", "van Halen", "McFly", "Wallace", "McLeod", "Skywalker", "Smith",
       "Silva", "da Silva", "Sexy", "Coupat", "Coupable", "Byron", "Lovelace", "Pascal", "Kareninski", "Dynamite",
       "Souza", "Ha", "Balboa", "Durden", "V.", "Li", "Manco", "Kelly", "Torquato", "Sampaio", "Bittencourt", "Parisi",
       "Oliveira", "Crap", "Copppercup", "Motherfucker", "Firehead", "Martin", "Papanicolau", "Galilei", "Stuart",
       "Bitch", "King", "Cleese", "Thoreau", "Twain", "Marx", "Yankovicz", "Vader", "Prado", "Teixeira", "Oliveira",
       "Nogueira", "Pereira", "Sant'anna", "Kerns", "Patel", "Ahmadzai", "Riding", "Llabjani", "Maus",
       "Liger", "Byrne", "Wood", "Angelov", "Andreu", "Sadeghi", "Gajjar", "Kara", "Wolstenholme", "Alghaith",
       "Young", "Scott", "Luz", "Copic", "Pucihar", "Zhou", "Dutta", "Baruah", "Singh", "Sauro", "do Nascimento",
       "Lee", "Trevisan", "Travisani", "Pereira", "Nandwani", "Moura", "Senna"]
_prefixes = ["Dr.", "Prof.", "Sir", "Ven."]
_suffixes = ["The 3rd", "Jr.", "Sobrinho", "Neto", "VIII", "XVI", "I", "II", "III", "IV"]
_PROB_PREF = 0.1
_PROB_SUFF = 0.1


def random_name(num_surnames=2):
    """
    Returns a random person name

    Arguments:
      num_surnames -- number of surnames
    """
    a = []

    # Prefix
    if random.random() < _PROB_PREF:
        a.append(_prefixes[random.randint(0, len(_prefixes) - 1)])

    # Forename
    a.append(_forenames[random.randint(0, len(_forenames) - 1)])

    # Surnames
    for i in range(num_surnames):
        a.append(_surnames[random.randint(0, len(_surnames) - 1)])

    # Suffix
    if random.random() < _PROB_SUFF:
        a.append(_suffixes[random.randint(0, len(_suffixes) - 1)])

    return " ".join(a)




# # todo cleanup
# # This is just for debugging
# # http://stackoverflow.com/questions/2023608/check-what-files-are-open-in-python
# import __builtin__
# openfiles = set()
# oldfile = __builtin__.file
# class newfile(oldfile):
#     def __init__(self, *args):
#         self.x = args[0]
#         print "### OPENING %s ###" % str(self.x)
#         oldfile.__init__(self, *args)
#         openfiles.add(self)
#
#     def close(self):
#         print "### CLOSING %s ###" % str(self.x)
#         oldfile.close(self)
#         openfiles.remove(self)
# oldopen = __builtin__.open
# def newopen(*args):
#     return newfile(*args)
# __builtin__.file = newfile
# __builtin__.open = newopen
#
# def printOpenFiles():
#     print "### %d OPEN FILES: [%s]" % (len(openfiles), ", ".join(f.x for f in openfiles))
# __all__.append("printOpenFiles")


# # Logger for internal use
# _logger = logging.getLogger(__name__)
# _logger.addHandler(logging.NullHandler())





"""Class & metaclass stuff"""


from functools import wraps
from collections import OrderedDict
from collections import defaultdict


__all__ = ["AttrsPart", "froze_it", "keydefaultdict"]



def froze_it(cls):
    """
    Decorator to prevent from creating attributes in the object ouside __init__().

    This decorator must be applied to the final class (doesn't work if a
    decorated class is inherited).

    Yoann's answer at http://stackoverflow.com/questions/3603502
    """
    cls._frozen = False

    def frozensetattr(self, key, value):
        if self._frozen and not hasattr(self, key):
            raise AttributeError("Attribute '{}' of class '{}' does not exist!"
                  .format(key, cls.__name__))
        else:
            object.__setattr__(self, key, value)

    def init_decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            func(self, *args, **kwargs)
            self._frozen = True
        return wrapper

    cls.__setattr__ = frozensetattr
    cls.__init__ = init_decorator(cls.__init__)

    return cls


class _AAObject(object):
    """
    Implements "meta" property (used to store history, etc.)

    TODO making use of this 'meta' is for the future
    """

    @property
    def classname(self):
        return self.__class__.__name__

    def __init__(self):
        self.meta = {}


class AttrsPart(_AAObject):
    """
    Implements a new __str__() to print selected attributes.

    **Subclassing**:
      - set the `attrs` class variable
    """

    # for __str__()
    attrs = None
    # for __repr__()
    # Optional; if not set, will be overwritten with self.attrs at __init__()
    less_attrs = None

    def __init__(self):
        _AAObject.__init__(self)
        if self.less_attrs is None:
            self.less_attrs = self.attrs

    def __str__(self):
        if self.attrs is None or len(self.attrs) == 0:
            return _AAObject.__str__(self)

        maxlen = max([len(x) for x in self.attrs])
        s_format = "{{:>{0:d}}} = {{}}".format(maxlen)
        l = []
        for x in self.attrs:
            y = self.__getattribute__(x)
            if isinstance(y, list):
                # list gets special treatment
                v = "["+\
                     ",\n{0:{1}}".format("", maxlen+4).join([z.one_liner_str() if isinstance(z, AttrsPart)
                                                   else str(z) for z in y])+\
                     "\n{0:{1}}]".format("", maxlen+3)
            else:
                v = self.__getattribute__(x)

            l.append(s_format.format(x, v))

        s = "\n".join(l)
        return s

    def one_liner_str(self):
        """Returns string (supposed to be) shorter than str() and not contain newline"""
        assert self.less_attrs is not None, "Forgot to set attrs class variable"
        s_format = "{}={}"
        s = "; ".join([s_format.format(x, self.__getattribute__(x)) for x in self.less_attrs])
        return s

    def to_dict(self):
        """Returns OrderedDict whose keys are self.attrs"""
        ret = OrderedDict()
        for attrname in self.attrs:
            ret[attrname] = self.__getattribute__(attrname)
        return ret

    def to_list(self):
        """Returns list containing values of attributes listed in self.attrs"""

        ret = OrderedDict()
        for attrname in self.attrs:
            ret[attrname] = self.__getattribute__(attrname)
        return ret


class keydefaultdict(defaultdict):
    """
    Subclass of defaultdict to pass the missing key to the default factory

    Usage:

        d = keydefaultdict(C)
        d[x] # returns C(x)

    Source: solution by Rochen Ritzel at https://stackoverflow.com/questions/2912231/is-there-a-clever-way-to-pass-the-key-to-defaultdicts-default-factory
    """

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError( key )
        else:
            ret = self[key] = self.default_factory(key)
            return ret




from PyQt5.QtWidgets import QMainWindow
from ._logpart import *

__all__ = ["XLogMainWindow"]

class XLogMainWindow(QMainWindow, _LogPart):
    """QMainWindow subclass that carries a logging facility"""
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        _LogPart.__init__(self)
        self._refs = []


import collections
import os.path
import re
import shutil
from threading import Lock
import sys
import a99
import logging


__all__ = [
    "rename_to_temp", "is_text_file", "multirow_str_vector", "add_bits_to_path", "crunch_dir",
    "slugify", "write_lf", "int_vector", "float_vector", "get_path",
    "str_vector", "new_filename", "readline_strip", "create_symlink", "which",
]


# # Filename or pathname-related string manipulations

def slugify(value, flagLower=True):
  """
  Converts to lowercase, removes non-alpha characters,
  and converts spaces to hyphens.

  Useful for making file names.

  Source: http://stackoverflow.com/questions/5574042/string-slugification-in-python
  """
  value = re.sub('[^\w\s.]', '', value).strip()
  if flagLower:
    value = value.lower()
  value = re.sub('[-\s]+', '-', value)
  return value


def crunch_dir(name, n=50):
    """Puts "..." in the middle of a directory name if lengh > n."""
    if len(name) > n + 3:
        name = "..." + name[-n:]
    return name

def add_bits_to_path(path_, filename_prefix=None, extension=None):
    """
    Adds prefix/suffix to filename

    Arguments:
        path_ -- path to file
        filename_prefix -- prefix to be added to file name
        extension -- extension to be added to file name. The dot is automatically added, such as
            "ext" and ".ext" will have the same effect

    Examples:
        > add_bits_to_path("/home/user/file", "prefix-")
        /home/user/prefix-file

        > add_bits_to_path("/home/user/file", None, ".ext")
        /home/user/file.ext

        > add_bits_to_path("/home/user/file", None, "ext")  # dot in extension is optional
        /home/user/file.ext

        > add_bits_to_path("/home/user/", None, ".ext")
        /home/user/.ext
    """

    dir_, basename = os.path.split(path_)

    if filename_prefix:
        basename = filename_prefix+basename
    if extension:
        if not extension.startswith("."):
            extension = "."+extension
        basename = basename+extension

    return os.path.join(dir_, basename)



# # Text file parsing utillities

def str_vector(f):
    """
    Reads next line of file and makes it a vector of strings

    Note that each str.split() already strips each resulting string of any whitespaces.
    """
    return f.readline().split()


def float_vector(f):
    """Reads next line of file and makes it a vector of floats."""
    return [float(s) for s in str_vector(f)]


def int_vector(f):
    """Reads next line of file and makes it a vector of floats."""
    return [int(s) for s in str_vector(f)]


def readline_strip(f):
    """Reads next line of file and strips the newline."""
    return f.readline().strip('\n')


def multirow_str_vector(f, n, r=0):
    """
    Assembles a vector that spans several rows in a text file.

    Arguments:
      f -- file-like object
      n -- number of values expected
      r (optional) -- Index of last row read in file (to tell which file row in
                      case of error)

    Returns:
      (list-of-strings, number-of-rows-read-from-file)
    """

    so_far = 0
    n_rows = 0
    v = []
    while True:
        temp = str_vector(f)
        n_rows += 1
        n_now = len(temp)

        if n_now+so_far > n:
            a99.get_python_logger().warning(('Reading multi-row vector: '
                'row %d should have %d values (has %d)') %
                (r+n_rows, n-so_far, n_now))

            v.extend(temp[:n-so_far])
            so_far = n

        elif n_now+so_far <= n:
            so_far += n_now
            v.extend(temp)

        if so_far == n:
            break

    return v, n_rows



# # Probe, write, rename etc.

def new_filename(prefix, extension=""):
    """returns a file name that does not exist yet, e.g. prefix.0001.extension

    If extension starts with an '.', this character will be ignored
    """

    if len(extension) > 0 and extension[0] == '.':
        extension = extension[1:]

    i = -1
    while True:
        if i > -1:
            ret = '{0!s}.{1:04d}.{2!s}'.format(prefix, i, extension) \
                if extension else '{0!s}.{1:04d}'.format(prefix, i)
        else:
            ret = "{}.{}".format(prefix, extension) \
                if extension else prefix

        if not os.path.exists(ret):
            break
        i += 1
        if i > 9999:
            raise RuntimeError("Could not make a new file name for (prefix='{0!s}', extension='{1!s}')".format(prefix, extension))
    return ret


_rename_to_temp_lock = Lock()
def rename_to_temp(filename):
    """*Thread-safe* renames file to temporary filename. Returns new name"""
    with _rename_to_temp_lock:
        root, ext = os.path.splitext(filename)
        if len(ext) > 0:
            ext = ext[1:]  # the dot (".") is originally included
        new_name = new_filename(root, ext)
        os.rename(filename, new_name)
        return new_name


def write_lf(h, s):
  """Adds lf to end of string and writes it to file."""
  h.write(s+"\n")


def create_symlink(source, link_name):
    """
    Creates symbolic link for either operating system.

    http://stackoverflow.com/questions/6260149/os-symlink-support-in-windows
    """
    os_symlink = getattr(os, "symlink", None)
    if isinstance(os_symlink, collections.Callable):
        os_symlink(source, link_name)
    else:
        import ctypes
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 1 if os.path.isdir(source) else 0
        if csl(link_name, source, flags) == 0:
            raise ctypes.WinError()





# ## http://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-lemented-in-python
_PY3 = sys.version_info[0] == 3

# A function that takes an integer in the 8-bit range and returns
# a single-character byte object in py3 / a single-character string
# in py2.
#
_int2byte = (lambda x: bytes((x,))) if _PY3 else chr

_text_characters = (
        b''.join(_int2byte(i) for i in range(32, 127)) +
        b'\n\r\t\f\b')

def is_text_file(filepath, blocksize=2**13):
    """ Uses heuristics to guess whether the given file is text or binary,
        by reading a single block of bytes from the file.
        If more than 30% of the chars in the block are non-text, or there
        are NUL ('\x00') bytes in the block, assume this is a binary file.
    """
    with open(filepath, "rb") as fileobj:
        block = fileobj.read(blocksize)
        if b'\x00' in block:
            # Files with null bytes are binary
            return False
        elif not block:
            # an empty file is considered a valid text file
            return True

        # Use translate's 'deletechars' argument to efficiently remove all
        # occurrences of _text_characters from the block
        nontext = block.translate(None, _text_characters)
        return float(len(nontext)) / len(block) <= 0.30



def which(program):
    """
    Mimics UNIX 'which' command: return full path to executable file

    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    """
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def get_path(*args, module=a99):
    """Returns full path to specified module

    Args:
      *args: are added at the end of module path with os.path.join()
      module: Python module, defaults to a99

    Returns: path string

    >>> get_path()
    """

    p = os.path.abspath(os.path.join(os.path.split(module.__file__)[0], *args))
    return p
